import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_horizontal_links(shape, np.ndarray[DTYPE_t, ndim=1] horizontal_links):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_links = n_rows * (n_cols - 1) + (n_rows - 1) * n_cols
    cdef int link_stride = 2 * n_cols - 1
    cdef int i, n

    i = 0
    for link in range(0, n_links, link_stride):
        for n in range(n_cols - 1):
            horizontal_links[i] = link + n
            i += 1


@cython.boundscheck(False)
def fill_vertical_links(shape, np.ndarray[DTYPE_t, ndim=1] vertical_links):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int link_stride = 2 * n_cols - 1
    cdef int n_links = n_rows * (n_cols - 1) + (n_rows - 1) * n_cols
    cdef int i, n

    i = 0
    for link in range(n_cols - 1, n_links, link_stride):
        for n in range(n_cols):
            vertical_links[i] = link + n
            i += 1


@cython.boundscheck(False)
def fill_patches_at_link(shape, np.ndarray[DTYPE_t, ndim=2] patches_at_link):
    cdef int link
    cdef int patch
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int patches_per_row = n_cols - 1
    cdef int n_links = (2 * n_cols - 1) * n_rows - n_cols

    # Interior horizontal links
    for row in range(1, n_rows - 1):
        patch = patches_per_row * (row - 1)
        link = (2 * n_cols - 1) * row
        for link in range(link, link + n_cols - 1):
            patches_at_link[link, 0] = patch
            patches_at_link[link, 1] = patch + patches_per_row
            patch += 1

    # Interior vertical links
    for row in range(0, n_rows - 1):
        patch = patches_per_row * row
        link = n_cols + (2 * n_cols - 1) * row
        for link in range(link, link + n_cols - 2):
            patches_at_link[link, 0] = patch + 1
            patches_at_link[link, 1] = patch
            patch += 1

    # Left edge
    patch = 0
    for link in range(n_cols - 1, n_links, 2 * n_cols - 1):
        patches_at_link[link, 0] = patch
        patches_at_link[link, 1] = - 1
        patch += patches_per_row

    # Right edge
    patch = patches_per_row - 1
    for link in range(2 * n_cols - 2, n_links, 2 * n_cols - 1):
        patches_at_link[link, 0] = - 1
        patches_at_link[link, 1] = patch
        patch += patches_per_row

    # Bottom edge
    patch = 0
    for link in range(n_cols - 1):
        patches_at_link[link, 0] = - 1
        patches_at_link[link, 1] = patch
        patch += 1

    # Top edge
    patch = patches_per_row * (n_rows - 2)
    for link in range(n_links - (n_cols - 1), n_links):
        patches_at_link[link, 0] = patch
        patches_at_link[link, 1] = - 1
        patch += 1


@cython.boundscheck(False)
def fill_nodes_at_link(shape, np.ndarray[DTYPE_t, ndim=2] nodes_at_link):
    cdef int row, col
    cdef int link
    cdef int node
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_links = (2 * n_cols - 1) * n_rows - n_cols
    cdef int links_per_row = 2 * n_cols - 1

    # Horizontal links
    for row in range(n_rows):
        node = row * n_cols
        link = row * links_per_row
        for col in range(n_cols - 1):
            nodes_at_link[link, 0] = node
            nodes_at_link[link, 1] = node + 1
            node += 1
            link += 1

    # Vertical links
    for row in range(0, n_rows - 1):
        node = row * n_cols
        link = row * links_per_row + n_cols - 1
        for col in range(n_cols):
            nodes_at_link[link, 0] = node
            nodes_at_link[link, 1] = node + n_cols
            node += 1
            link += 1



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.OverlandFlow

last updated: 3/14/16
"""
import numpy as np

from landlab import RasterModelGrid
from landlab.components.overland_flow import OverlandFlow
from landlab.grid.structured_quad.links import left_edge_horizontal_ids


(_SHAPE, _SPACING, _ORIGIN) = ((32, 240), (25, 25), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_deAlm_name(deAlm):
    assert deAlm.name == 'OverlandFlow'


def test_deAlm_input_var_names(deAlm):
    assert deAlm.input_var_names == (
        'surface_water__depth',
        'topographic__elevation',
    )


def test_deAlm_output_var_names(deAlm):
    assert deAlm.output_var_names == (
        'surface_water__depth',
        'surface_water__discharge',
        'water_surface__gradient',
    )

def test_deAlm_var_units(deAlm):
    assert set(deAlm.input_var_names) | set(deAlm.output_var_names) == set(dict(deAlm.units).keys())

    assert deAlm.var_units('surface_water__depth') == 'm'
    assert deAlm.var_units('surface_water__discharge') == 'm3/s'
    assert deAlm.var_units('water_surface__gradient') == '-'
    assert deAlm.var_units('topographic__elevation') == 'm'


def test_grid_shape(deAlm):
    assert deAlm.grid.number_of_node_rows == _SHAPE[0]
    assert deAlm.grid.number_of_node_columns == _SHAPE[1]


def test_deAlm_analytical():
    grid = RasterModelGrid((32, 240), spacing = 25)
    grid.add_zeros('node', 'surface_water__depth')
    grid.add_zeros('node', 'topographic__elevation')
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0

    while time < 500.:
        grid.at_link['surface_water__discharge'][left_inactive_ids] = (
            grid.at_link['surface_water__discharge'][left_inactive_ids + 1])
        dt = deAlm.calc_time_step()
        deAlm.overland_flow(dt)
        h_boundary = (((7./3.) * (0.01**2) * (0.4**3) *
                      time) ** (3./7.))
        grid.at_node['surface_water__depth'][grid.nodes[1: -1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = (-(7./3.) * (0.01**2) * (0.4**2) * (x - (0.4 * 500)))

    h_analytical[np.where(h_analytical > 0)] = (h_analytical[np.where(
        h_analytical > 0)] ** (3./7.))
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)


def test_deAlm_analytical_imposed_dt_short():
    grid = RasterModelGrid((32, 240), spacing=25)
    grid.add_zeros('node', 'surface_water__depth')
    grid.add_zeros('node', 'topographic__elevation')
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0

    while time < 500.:
        grid.at_link['surface_water__discharge'][left_inactive_ids] = (
            grid.at_link['surface_water__discharge'][left_inactive_ids + 1])
        dt = 10.
        deAlm.overland_flow(dt)
        h_boundary = (((7./3.) * (0.01**2) * (0.4**3) *
                      time) ** (3./7.))
        grid.at_node['surface_water__depth'][grid.nodes[1: -1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = (-(7./3.) * (0.01**2) * (0.4**2) * (x - (0.4 * 500)))

    h_analytical[np.where(h_analytical > 0)] = (h_analytical[np.where(
        h_analytical > 0)] ** (3./7.))
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)


def test_deAlm_analytical_imposed_dt_long():
    grid = RasterModelGrid((32, 240), spacing=25)
    grid.add_zeros('node', 'surface_water__depth')
    grid.add_zeros('node', 'topographic__elevation')
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    left_inactive_ids = left_edge_horizontal_ids(grid.shape)
    deAlm = OverlandFlow(grid, mannings_n=0.01, h_init=0.001)
    time = 0.0

    while time < 500.:
        grid.at_link['surface_water__discharge'][left_inactive_ids] = (
            grid.at_link['surface_water__discharge'][left_inactive_ids + 1])
        dt = 100.
        deAlm.run_one_step(dt)
        h_boundary = (((7./3.) * (0.01**2) * (0.4**3) *
                      time) ** (3./7.))
        grid.at_node['surface_water__depth'][grid.nodes[1: -1, 1]] = h_boundary
        time += dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = (-(7./3.) * (0.01**2) * (0.4**2) * (x - (0.4 * 500)))

    h_analytical[np.where(h_analytical > 0)] = (h_analytical[np.where(
        h_analytical > 0)] ** (3./7.))
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hdeAlm = deAlm.h.reshape(grid.shape)
    hdeAlm = hdeAlm[1][1:]
    hdeAlm = np.append(hdeAlm, [0])
    np.testing.assert_almost_equal(h_analytical, hdeAlm, decimal=1)



#! /usr/bin/env python
"""
Created on Thu Jun 29 11:02:28 2017

@author: njlyons
"""

from landlab import imshow_grid, RasterModelGrid
from landlab.plot.event_handler import query_grid_on_button_press
from matplotlib.pyplot import gcf
from matplotlib.backend_bases import Event
from numpy.testing import assert_equal


def test_query_grid_on_button_press():
    
    rmg = RasterModelGrid((5, 5))
    imshow_grid(rmg, rmg.nodes, cmap='RdYlBu')
    
    # Programmatically create an event near the grid center.
    event = Event('simulated_event', gcf().canvas)
    event.xdata = int(rmg.number_of_node_columns * 0.5)
    event.ydata = int(rmg.number_of_node_rows * 0.5)
    
    results = query_grid_on_button_press(event, rmg)
    x_coord = results['grid location']['x_coord']
    y_coord = results['grid location']['x_coord']
    
    msg = 'Items: Simulated matplotlib event and query results.'
    assert_equal(x_coord, event.xdata, msg)
    assert_equal(y_coord, event.ydata, msg)
    
    msg = 'Items: Node ID and grid coordinates of simulated matplotlib event.'
    node = rmg.grid_coords_to_node_id(event.xdata, event.ydata)
    assert_equal(results['node']['ID'], node, msg)



# -*- coding: utf-8 -*-
"""
Created on Wed Nov  4 08:10:22 2015

@author: gtucker

List of needs:

grid.link_at_face: IDs of links at each face
gradient functions:
    - of a node value at faces
    - of a node value at links
    - of a node value at junctions
a consistent way to sort
    - nodes: by y then x
    - links and faces: by junction coords (y then x)
    
    
Notes:

    grads_at_links:
        lg = (nv[node_at_link_head]-nv[node_at_link_tail])/link_length
    grads_at_faces:
        fg = lg[link_at_face]
        OR
        fg = (nv[node_at_face_right]-nv[node_at_face_left])/link_length[link_at_face]
        OR
        fg = (nv[node_at_face_right]-nv[node_at_face_left])/link_length_face

CURRENT STICKING POINT, 11/6/15:
WANT TO IMPLEMENT THE 3 VERSIONS ABOVE. BUT NO PROPER FACE-LINK DATA STRUCTURE,
AND CELL AREAS SEEMS TO BE BROKEN...
RELATES TO THE NEED TO IDENTIFY PERIMETER NODES IN A VORONOI GRID.
IN A RECTANGULAR RASTER, NODES WITH CELLS ARE EASY TO IDENTIFY.
IN A HEX-SHAPED OR RECTANGULAR HEX, SLIGHTLY MORE COMPLICATED BUT NOT TOO BAD.
IN A GENERAL VORONOI...??
FIRST THING PROBABLY IS TO FIX LINK_AT_FACE FOR RASTER, THEN IMPLEMENT FOR HEX.
"""
from six.moves import range

from landlab import HexModelGrid
import numpy as np
from numpy.testing import assert_array_equal
import time

MAX_NUM_LINKS = 6


def gt_grads_at_faces1(grid, nv):
    
    lg = (nv[grid.node_at_link_head]-nv[grid.node_at_link_tail])/grid.length_of_link
    #return lg[grid.link_at_face]
    return None # temporary
    

def gt_link_flux_divergence_at_cells_with_2darray(grid, f, out=None):
    
    if out is not None:
        net_flux = out
    else:
        net_flux = np.zeros(grid.number_of_nodes)
        
    for r in range(MAX_NUM_LINKS):
        links = grid.gt_links_at_node[r,grid.node_at_cell]
        net_flux += f[links]*grid.gt_link_dirs_at_node[r,:]


def gt_calc_gradients_at_faces(grid, vn):
    # vn = values at nodes
    print(grid.link_at_face)


def make_links_at_node_array(grid):
    """Make array with links at each node"""
    
    # Create arrays for link-at-node information
    grid.gt_links_at_node = -np.ones((MAX_NUM_LINKS, grid.number_of_nodes), dtype=np.int32)
    grid.gt_link_dirs_at_node = np.zeros((MAX_NUM_LINKS, grid.number_of_nodes), dtype=np.int8)
    grid.gt_active_link_dirs_at_node = np.zeros((MAX_NUM_LINKS, grid.number_of_nodes), dtype=np.int8)
    grid.gt_num_links_at_node = np.zeros(grid.number_of_nodes, dtype=np.uint8)  # assume <256 links at any node
    grid.gt_num__active_links_at_node = np.zeros(grid.number_of_nodes, dtype=np.uint8)  # assume <256 links at any node
    
    # Sweep over all links
    for lk in range(grid.number_of_links):
        
        # Find the ID of the tail node
        t = grid.node_at_link_tail[lk]
        
        # Its row in the 2D array is equal to the number of links we've found
        # so far for this node
        index = grid.gt_num_links_at_node[t]
        
        # Add this link to the list for this node, set the direction (outgoing,
        # indicated by -1), and increment the number found so far
        grid.gt_links_at_node[index][t] = lk
        grid.gt_link_dirs_at_node[index][t] = -1
        grid.gt_num_links_at_node[t] += 1
        
        # Find the ID of the tail node
        h = grid.node_at_link_head[lk]
        
        # Its row in the 2D array is equal to the number of links we've found
        # so far for this node
        index = grid.gt_num_links_at_node[h]
        
        # Add this link to the list for this node, set the direction
        # (incoming, indicated by +1), and increment the number found so far
        grid.gt_links_at_node[index][h] = lk
        grid.gt_link_dirs_at_node[index][h] = 1
        grid.gt_num_links_at_node[h] += 1
        

def testing_flux_divergence_with_hex():
    """Test flux divergence function(s).

    Notes
    -----
    Test grid looks like this:

        (7)-17-.(8)-18-.(9)
        . .     . .     . .
      11  12  13  14  15  16
      /     \ /     \ /     \ 
    (3)--8-.(4)--9-.(5)-10-.(6)
      .     . .     . .     .
       2   3   4   5   6   7
        \ /     \ /     \ /
        (0)--0-.(1)--1-.(2)

    Node numbers in parentheses; others are link numbers; period indicates
    link head.
    """
    hmg = HexModelGrid(3, 3, reorient_links=True)
    
    f = hmg.add_zeros('link', 'test_flux')
    f[:] = np.arange(hmg.number_of_links)

    make_links_at_node_array(hmg)
    
    assert_array_equal(hmg.gt_num_links_at_node,
                       [3, 4, 3, 3, 6, 6, 3, 3, 4, 3])
                       
    assert_array_equal(hmg.gt_links_at_node,
                       [[  0,  0,  1,  2,  3,  5,  7, 11, 13, 15],
                        [  2,  1,  6,  8,  4,  6, 10, 12, 14, 16],
                        [  3,  4,  7, 11,  8,  9, 16, 17, 17, 18],
                        [ -1,  5, -1, -1,  9, 10, -1, -1, 18, -1],
                        [ -1, -1, -1, -1, 12, 14, -1, -1, -1, -1],
                        [ -1, -1, -1, -1, 13, 15, -1, -1, -1, -1]])
                       
    assert_array_equal(hmg.gt_link_dirs_at_node,
                       [[ -1,  1,  1,  1,  1,  1,  1,  1,  1,  1],
                        [ -1, -1, -1, -1,  1,  1,  1,  1,  1,  1],
                        [ -1, -1, -1, -1,  1,  1, -1, -1,  1,  1],
                        [  0, -1,  0,  0, -1, -1,  0,  0, -1,  0],
                        [  0,  0,  0,  0, -1, -1,  0,  0,  0,  0],
                        [  0,  0,  0,  0, -1, -1,  0,  0,  0,  0]])

    raw_net_flux = np.zeros(hmg.number_of_nodes)
    for r in range(MAX_NUM_LINKS):
        raw_net_flux += f[hmg.gt_links_at_node[r,:]]*hmg.gt_link_dirs_at_node[r,:]
    assert_array_equal(raw_net_flux, 
                       [ -5., -10., -12., -17., -19., -19.,  1.,  6., 26., 49.])

    nv = np.arange(hmg.number_of_nodes)
    #gt_calc_gradients_at_faces(hmg, nv)    
    #gt_link_flux_divergence_at_cells_with_2darray(hmg, f)
    
    # Some time trials
    start = time.time()
    for i in range(1000):
        gt_grads_at_faces1(hmg, nv)
    endtime = time.time()


if __name__=='__main__':
    testing_flux_divergence_with_hex()



# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.overland_flow.OverlandFlowBates

last updated: 3/14/16
"""
import numpy as np

from landlab import RasterModelGrid
from landlab.components.overland_flow import OverlandFlowBates, OverlandFlow


(_SHAPE, _SPACING, _ORIGIN) = ((32, 240), (25, 25), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_Bates_name(bates):
    assert bates.name == 'OverlandFlowBates'


def test_Bates_input_var_names(bates):
    assert set(bates.input_var_names) == set(('surface_water__depth', 'topographic__elevation'))


def test_Bates_output_var_names(bates):
    assert set(bates.output_var_names) == set(
        (
            'surface_water__depth',
            'surface_water__discharge',
            'water_surface__gradient',
        )
    )

def test_Bates_var_units(bates):
    assert set(bates.input_var_names) | set(bates.output_var_names) == set(dict(bates.units).keys())

    assert bates.var_units('surface_water__depth') == 'm'
    assert bates.var_units('surface_water__discharge') == 'm3/s'
    assert bates.var_units('water_surface__gradient') == 'm/m'
    assert bates.var_units('topographic__elevation') == 'm'


def test_field_initialized_to_zero(bates):
    for name in bates.grid['node'].keys():
        field = bates.grid['node'][name]
        if name != 'surface_water__depth':
            assert np.all(np.isclose(field, 0.))
        else:
            assert np.all(np.isclose(field, 0.001))
            # this remains broken, and needs JA's attention


def test_grid_shape(bates):
    assert bates.grid.number_of_node_rows == _SHAPE[0]
    assert bates.grid.number_of_node_columns == _SHAPE[1]


def test_Bates_analytical():
    from landlab import RasterModelGrid
    grid = RasterModelGrid((32, 240), spacing = 25)
    grid.add_zeros('node', 'surface_water__depth')
    grid.add_zeros('node', 'topographic__elevation')
    grid.set_closed_boundaries_at_grid_edges(True, True, True, True)
    bates = OverlandFlowBates(grid, mannings_n = 0.01, h_init=0.001)
    time = 0.0
    bates.dt = 1.0
    while time < 500:
        bates.overland_flow(grid)
        h_boundary = (((7./3.) * (0.01**2) * (0.4**3) *
                  time) ** (3./7.))
        grid.at_node['surface_water__depth'][grid.nodes[1: -1, 1]] = h_boundary
        time += bates.dt

    x = np.arange(0, ((grid.shape[1]) * grid.dx), grid.dx)
    h_analytical = (-(7./3.) * (0.01**2) * (0.4**2) * (x - (0.4 * 500)))

    h_analytical[np.where(h_analytical > 0)] = (h_analytical[np.where(
        h_analytical > 0)] ** (3./7.))
    h_analytical[np.where(h_analytical < 0)] = 0.0

    hBates = bates.h.reshape(grid.shape)
    hBates = hBates[1][1:]
    hBates = np.append(hBates, [0])
    np.testing.assert_almost_equal(h_analytical, hBates, decimal=1)



import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_unit_spacing():
    """Test on a grid with unit spacing."""
    grid = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    grads = grid.calc_grad_at_link(values_at_nodes)
    assert_array_equal(
        grads,
        np.array([1, 1, 1, 1, 
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1],
                 dtype=float))
    diffs = grid.calculate_diff_at_links(values_at_nodes)
    assert_array_equal(grads, diffs)


def test_non_unit_spacing():
    """Test on a grid with non-unit spacing."""
    grid = RasterModelGrid((4, 5), spacing=(5, 2))
    values_at_nodes = np.arange(20.)
    grads = grid.calc_grad_at_link(values_at_nodes)
    assert_array_equal(
        grads,
        np.array(
           [0.5, 0.5,  0.5,  0.5,
            1.0, 1.0,  1. ,  1. ,  1. ,  
            0.5,  0.5, 0.5,  0.5,
            1. ,  1. , 1. ,  1. ,  1. ,
            0.5,  0.5,  0.5,  0.5,
            1. ,  1. ,  1. ,  1. , 1. ,
            0.5,  0.5,  0.5,  0.5],
            dtype=float))
    diffs = grid.calc_diff_at_link(values_at_nodes)
    assert_array_equal(diffs,
        np.array([1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1,
                  5, 5, 5, 5, 5,
                  1, 1, 1, 1],
                 dtype=float))


def test_out_array():
    """Test using the out keyword."""
    grid = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    grads = np.empty(31)
    rtn_grads = grid.calc_grad_at_link(values_at_nodes, out=grads)
    assert_array_equal(
        grads,
        np.array([1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1, 5, 5,
                  5, 5, 5, 1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1],
                 dtype=float))
    assert rtn_grads is grads


def test_diff_out_array():
    """Test that return array is the out keyword."""
    grid = RasterModelGrid((4, 5))
    values_at_nodes = np.arange(20.)
    diff = np.empty(31)
    rtn_diff = grid.calculate_diff_at_links(values_at_nodes, out=diff)
    assert_array_equal(
        diff,
        np.array([1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1, 5, 5,
                  5, 5, 5, 1, 1, 1, 1, 5, 5, 5, 5, 5, 1, 1, 1, 1],
                 dtype=float))
    assert rtn_diff is diff



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Unit tests for KinwaveImplicitOverlandFlowModel.

Created on Sat Apr  1 10:49:33 2017

@author: gtucker
"""

from landlab import RasterModelGrid
from landlab.components import KinwaveImplicitOverlandFlow
from numpy.testing import assert_array_equal
import numpy as np


def test_initialization():
    """Test initialization with various parameters.
    """
    rg = RasterModelGrid((3, 4), 2.0)
    rg.add_zeros('node', 'topographic__elevation')
    kw = KinwaveImplicitOverlandFlow(rg)

    # Make sure fields have been created
    for field_name in kw._var_mapping:
        if kw._var_mapping[field_name] == 'node':
            assert field_name in kw.grid.at_node
        elif kw._var_mapping[field_name] == 'link':
            assert field_name in kw.grid.at_link

    # Re-initialize, this time with fields already existing in the grid
    # (this triggers the "if" instead of "else" in the field setup in init)
    kw = KinwaveImplicitOverlandFlow(rg)


def test_first_iteration():
    """Test stuff that happens only on first iteration"""
    
    # Create a basic ramp
    rg = RasterModelGrid((10,10), spacing=(2, 2))
    rg.add_field('topographic__elevation', 0.1 * rg.node_y, at='node')

    # Create component and run it
    kw = KinwaveImplicitOverlandFlow(rg)
    kw.run_one_step(1.0)

    # Max gradient should be 0.1, and min should be zero
    assert round(np.amax(kw.grid.at_link['topographic__gradient']), 2) == 0.1
    assert round(np.amin(kw.grid.at_link['topographic__gradient']), 2) == 0.0
    assert round(np.amax(kw.sqrt_slope), 3) == 0.316
    assert round(np.amax(kw.grad_width_sum), 3) == 0.632
    assert round(np.amax(kw.alpha), 3) == 15.811


def test_steady_basic_ramp():
    """Run to steady state with basic ramp"""

    # Create a basic ramp
    rg = RasterModelGrid((10,10), spacing=(2, 2))
    rg.add_field('topographic__elevation', 0.1 * rg.node_y, at='node')

    # Create component and run it
    kw = KinwaveImplicitOverlandFlow(rg)
    for i in range(12):
        kw.run_one_step(1.0, runoff_rate=0.001)

    # Look at a column of nodes down the middle. The inflow from uphill should
    # be, from top to bottom: 0, 0.004, 0.008, 0.012, 0.016, 0.02, 0.024, 0.028
    assert kw.disch_in[85] == 0.0
    assert round(kw.disch_in[75], 3) == 0.004
    assert round(kw.disch_in[65], 3) == 0.008
    assert round(kw.disch_in[55], 3) == 0.012
    assert round(kw.disch_in[45], 3) == 0.016
    assert round(kw.disch_in[35], 3) == 0.020
    assert round(kw.disch_in[25], 3) == 0.024
    assert round(kw.disch_in[15], 3) == 0.028

    # Try with passing in runoff
    kw = KinwaveImplicitOverlandFlow(rg, runoff_rate=360.0)
    kw.depth[:] = 0.0
    for i in range(22):
        kw.run_one_step(1.0)

    # Again, look at a column of nodes down the middle. The inflow from uphill 
    # should now be 1/10 of the prior example.
    assert round(kw.disch_in[75], 4) == 0.0004
    assert round(kw.disch_in[65], 4) == 0.0008
    assert round(kw.disch_in[55], 4) == 0.0012
    assert round(kw.disch_in[45], 4) == 0.0016
    assert round(kw.disch_in[35], 4) == 0.0020
    assert round(kw.disch_in[25], 4) == 0.0024
    assert round(kw.disch_in[15], 4) == 0.0028

    # Try with default runoff rate of 1 mm/hr = 2.78e-7 m/s
    kw = KinwaveImplicitOverlandFlow(rg)
    assert round(kw.runoff_rate * 1.0e7, 2) == 2.78
    kw.depth[:] = 0.0
    for i in range(18):
        kw.run_one_step(10.0)

    # Look at a column of nodes down the middle. The inflow from uphill should
    # be, from top to bottom: 0, 0.004, 0.008, 0.012, 0.016, 0.02, 0.024, 0.028
    assert kw.disch_in[85] == 0.0
    assert round(kw.disch_in[75], 7) == 0.0000011
    assert round(kw.disch_in[65], 7) == 0.0000022
    assert round(kw.disch_in[55], 7) == 0.0000033
    assert round(kw.disch_in[45], 7) == 0.0000044
    assert round(kw.disch_in[35], 7) == 0.0000055
    assert round(kw.disch_in[25], 7) == 0.0000066
    assert round(kw.disch_in[15], 7) == 0.0000077


def test_curved_surface():
    """Test flow across a curved surface."""

    # Create a grid
    rg = RasterModelGrid((10,10), spacing=(2, 2))
    rg.add_field('topographic__elevation', 3.*rg.node_x**2 + rg.node_y**2,
                 at='node')

    # Create component and run it
    kw = KinwaveImplicitOverlandFlow(rg)
    for i in range(8):
        kw.run_one_step(1.0, runoff_rate=0.001)

    # The inflow discharge to each cell at steady state should equal the
    # runoff rate times the "inflow" drainage area, which is the total drainage
    # area minus the area of the cell itself. Here we'll test a column of core
    # nodes across the middle of the domain.
    area = rg.at_node['drainage_area']
    runoff_rate = 0.001
    unit_area = 4.0
    for i in range(15, 95, 10):
        assert round(kw.disch_in[i], 6) == round(runoff_rate * (area[i] - unit_area), 6)


if __name__ == '__main__':
    test_initialization()
    test_first_iteration()
    test_steady_basic_ramp()
    test_curved_surface()



"""Test simple stream power functionality when a width array is specified."""
import os

from six.moves import range

import numpy
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from landlab import RasterModelGrid
from landlab import ModelParameterDictionary
from landlab.components.flow_routing import FlowRouter
from landlab.components.stream_power import StreamPowerEroder


_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_sp_widths():
    input_str = os.path.join(_THIS_DIR, 'test_sp_params_widths.txt')
    inputs = ModelParameterDictionary(input_str, auto_type=True)
    nrows = 5
    ncols = 5
    dx = inputs.read_float('dx')
    dt = inputs.read_float('dt')

    mg = RasterModelGrid(nrows, ncols, dx)
    widths = np.ones(mg.number_of_nodes, dtype=float)
    mg.add_zeros('topographic__elevation', at='node')
    z = np.array([5., 5., 0., 5., 5.,
                  5., 2., 1., 2., 5.,
                  5., 3., 2., 3., 5.,
                  5., 4., 4., 4., 5.,
                  5., 5., 5., 5., 5.])
    mg['node']['topographic__elevation'] = z

    fr = FlowRouter(mg)
    sp = StreamPowerEroder(mg, use_W=widths, **inputs)

    # perform the loop (once!)
    for i in range(1):
        fr.route_flow()
        sqrt_A = mg.at_node['drainage_area']**0.5
        widths[mg.core_nodes] = sqrt_A[mg.core_nodes]/sqrt_A[
            mg.core_nodes].mean()
        # so widths has mean=1.
        # note the issue with drainage_area not defined at perimeter => nans
        # if not careful...
        sp.run_one_step(dt)

    z_tg = np.array([ 5.        ,  5.        ,  0.        ,  5.        ,
                      5.        ,  5.        ,  1.37222369,  0.36876358,
                      1.37222369,  5.        ,  5.        ,  2.17408606,
                      1.07986038,  2.17408606,  5.        ,  5.        ,
                      3.08340277,  2.85288049,  3.08340277,  5.        ,
                      5.        ,  5.        ,  5.        ,  5.        ,
                      5.        ])

    assert_array_almost_equal(mg.at_node['topographic__elevation'], z_tg)


# -*- coding: utf-8 -*-
"""
Unit tests for landlab.components.soil_moisture.SoilInfiltrationGreenAmpt

last updated: 3/14/16
"""
import numpy as np

from landlab import RasterModelGrid
from landlab.components.soil_moisture import SoilInfiltrationGreenAmpt


(_SHAPE, _SPACING, _ORIGIN) = ((10, 10), (25, 25), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_SI_name(si):
    assert si.name == 'SoilInfiltrationGreenAmpt'


def test_SI_input_var_names(si):
    assert si.input_var_names == ('surface_water__depth', 'soil_water_infiltration__depth',)


def test_SI_output_var_names(si):
    assert si.output_var_names == ('surface_water__depth', 'soil_water_infiltration__depth',)


def test_SI_var_units(si):
    assert set(si.input_var_names) | set(si.output_var_names) == set(dict(si.units).keys())

    assert si.var_units('surface_water__depth') == 'm'
    assert si.var_units('soil_water_infiltration__depth') == 'm'


def test_grid_shape(si):
    assert si.grid.number_of_node_rows == _SHAPE[0]
    assert si.grid.number_of_node_columns == _SHAPE[1]


def test_calc_soil_pressure(si):
    np.testing.assert_almost_equal(si.calc_soil_pressure('silt loam'),
                                   0.1647870740305523, decimal=6)

def test_calc_soil_head(si):
    soil_props = SoilInfiltrationGreenAmpt.SOIL_PROPS['loam']
    np.testing.assert_almost_equal(si.calc_pressure_head(soil_props[0],
            soil_props[1]), 0.087498292, decimal=6)


def test_calc_moisture_deficit(si):
    np.testing.assert_almost_equal(si.calc_moisture_deficit(
        soil_bulk_density=1700., rock_density=2650.,
        volume_fraction_coarse_fragments=0.,
        soil_moisture_content=0.2), 0.15849056603, decimal=6)


def test_run_one_step():
    grid = RasterModelGrid((10, 10), spacing=25)
    grid.add_ones('node', 'soil_water_infiltration__depth', dtype=float)
    grid.add_ones('node', 'surface_water__depth')
    hydraulic_conductivity = (2.5 *  (10**-6))
    grid['node']['surface_water__depth'] *= 5.0
    grid['node']['soil_water_infiltration__depth'] *= (10**-5)
    SI = SoilInfiltrationGreenAmpt(grid,
        hydraulic_conductivity=hydraulic_conductivity,
         soil_bulk_density=1700., rock_density=2650.,
         initial_soil_moisture_content=0.2, soil_type='silt loam',
         volume_fraction_coarse_fragments=0.6,
         coarse_sed_flag=False,
         surface_water_minimum_depth=1.e-7,
         soil_pore_size_distribution_index=None,
         soil_bubbling_pressure=None,
         wetting_front_capillary_pressure_head=None)

    SI.run_one_step(dt=5)
    np.testing.assert_almost_equal(grid['node']['surface_water__depth'][0],
                3.97677483519, decimal=6)



#! /usr/bin/env python
"""Unit tests for landlab.components.flexure.Flexure1D."""
import pytest
from numpy.testing import (
    assert_array_equal,
    assert_array_less,
    assert_array_almost_equal,
    assert_almost_equal,
)

import numpy as np

from landlab import RasterModelGrid
from landlab.components.flexure import Flexure1D


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e3, 10e3), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(flex1d):
    """Test component name exists and is a string."""
    assert isinstance(flex1d.name, str)


def test_input_var_names(flex1d):
    """Test input_var_names is a tuple of strings."""
    assert isinstance(flex1d.input_var_names, tuple)
    for name in flex1d.input_var_names:
        assert isinstance(name, str)


def test_output_var_names(flex1d):
    """Test output_var_names is a tuple of strings."""
    assert isinstance(flex1d.output_var_names, tuple)
    for name in flex1d.output_var_names:
        assert isinstance(name, str)


def test_var_units(flex1d):
    """Test input/output var units."""
    assert isinstance(flex1d.units, tuple)
    for name, units in flex1d.units:
        assert name in flex1d.input_var_names + flex1d.output_var_names
        assert isinstance(units, str)


def test_var_mapping(flex1d):
    """Test input/output var mappings."""
    assert isinstance(flex1d._var_mapping, dict)
    for name in flex1d.input_var_names + flex1d.output_var_names:
        assert name in flex1d._var_mapping
        assert isinstance(flex1d._var_mapping[name], str)
        assert flex1d._var_mapping[name] in (
            "node",
            "link",
            "patch",
            "corner",
            "face",
            "cell",
        )


def test_var_doc(flex1d):
    """Test input/output var docs."""
    assert isinstance(flex1d._var_doc, dict)
    for name in flex1d.input_var_names + flex1d.output_var_names:
        assert name in flex1d._var_doc
        assert isinstance(flex1d._var_doc[name], str)


def test_calc_airy():
    """Test airy isostasy."""
    flex = Flexure1D(RasterModelGrid((3, 5)), method="airy")
    flex.load_at_node[:] = flex.gamma_mantle

    assert_array_equal(flex.dz_at_node, 0.)
    flex.update()
    assert_array_equal(flex.dz_at_node, 1.)


def test_run_one_step():
    """Test the run_one_step method."""
    flex = Flexure1D(RasterModelGrid((3, 5)), method="airy")
    flex.load_at_node[:] = flex.gamma_mantle

    assert_array_equal(flex.dz_at_node, 0.)
    flex.run_one_step()
    assert_array_equal(flex.dz_at_node, 1.)


def test_with_one_row():
    """Test calculating on one row."""
    flex = Flexure1D(RasterModelGrid((3, 5)), method="airy", rows=1)
    flex.load_at_node[:] = flex.gamma_mantle

    assert_array_equal(flex.dz_at_node, 0.)
    flex.update()
    assert_array_equal(flex.dz_at_node[0], 0.)
    assert_array_equal(flex.dz_at_node[1], 1.)
    assert_array_equal(flex.dz_at_node[2], 0.)


def test_with_one_row():
    """Test calculating on one row."""
    flex = Flexure1D(RasterModelGrid((3, 5)), method="airy", rows=(0, 2))
    flex.load_at_node[:] = -flex.gamma_mantle

    assert_array_equal(flex.dz_at_node, 0.)
    flex.update()
    assert_array_equal(flex.dz_at_node[0], -1.)
    assert_array_equal(flex.dz_at_node[1], 0.)
    assert_array_equal(flex.dz_at_node[2], -1.)


def test_field_is_updated():
    """Test the output field is updated."""
    flex = Flexure1D(RasterModelGrid((3, 5)), method="airy", rows=(0, 2))
    flex.load_at_node[:] = -flex.gamma_mantle

    assert_array_equal(flex.dz_at_node, 0.)
    flex.update()

    dz = flex.grid.at_node["lithosphere_surface__increment_of_elevation"]
    assert_array_equal(flex.dz_at_node.flatten(), dz)


def test_calc_flexure():
    """Test calc_flexure function."""
    x = np.arange(100.)
    loads = np.ones(100)
    dz = Flexure1D.calc_flexure(x, loads, 1., 1.)

    assert_array_less(0., dz)
    assert isinstance(dz, np.ndarray)
    assert dz.shape == loads.shape
    assert dz.dtype == loads.dtype


def test_calc_flexure_with_out_keyword():
    """Test calc_flexure out keyword."""
    x = np.arange(100.)
    loads = np.ones(100)
    buffer = np.empty_like(x)
    dz = Flexure1D.calc_flexure(x, loads, 1., 1., out=buffer)
    assert np.may_share_memory(dz, buffer)


def test_calc_flexure():
    """Test calc_flexure with multiple rows of loads."""
    x = np.arange(100.) * 1e3
    loads = np.ones(500).reshape((5, 100))
    dz = Flexure1D.calc_flexure(x, loads, 1e4, 1.)

    assert_array_less(0., dz)
    assert isinstance(dz, np.ndarray)
    assert dz.shape == loads.shape
    assert dz.dtype == loads.dtype

    for row in range(5):
        assert_array_almost_equal(dz[0], dz[row])


DEPENDS_ON = {
    "eet": ("rigidity", "alpha"),
    "youngs": ("rigidity", "alpha"),
    "rho_water": ("gamma_mantle", "alpha"),
    "rho_mantle": ("gamma_mantle", "alpha"),
    "gravity": ("gamma_mantle", "alpha"),
}


def test_setters(flexure_keyword):
    EPS = 1e-6
    flex = Flexure1D(RasterModelGrid((3, 5)))
    val_before = {}
    for name in DEPENDS_ON[flexure_keyword]:
        val_before[name] = 1. * getattr(flex, name)
    for name in DEPENDS_ON[flexure_keyword]:
        setattr(
            flex, flexure_keyword, getattr(flex, flexure_keyword) * (1. + EPS) + EPS
        )
        assert val_before[name] != getattr(flex, name)


def test_method_keyword():
    """Test using the method keyword."""
    flex = Flexure1D(RasterModelGrid((3, 5)), method="airy")
    assert flex.method == "airy"
    flex = Flexure1D(RasterModelGrid((3, 5)), method="flexure")
    assert flex.method == "flexure"
    with pytest.raises(ValueError):
        Flexure1D(RasterModelGrid((3, 5)), method="Flexure")


def test_flexure_keywords(flexure_keyword):
    flex = Flexure1D(RasterModelGrid((3, 5)), **{flexure_keyword: 1.})
    assert getattr(flex, flexure_keyword) == 1.
    assert isinstance(getattr(flex, flexure_keyword), float)
    with pytest.raises(ValueError):
        flex = Flexure1D(RasterModelGrid((3, 5)), **{flexure_keyword: -1})


def test_x_at_node():
    """Test x_at_node is reshaped and shares memory with the grid."""
    flex = Flexure1D(RasterModelGrid((3, 5)))

    assert_array_equal(
        flex.x_at_node,
        [[0., 1., 2., 3., 4.], [0., 1., 2., 3., 4.], [0., 1., 2., 3., 4.]],
    )


def test_dz_at_node():
    """Test dz_at_node is reshaped and shares memory with its field."""
    flex = Flexure1D(RasterModelGrid((3, 5)))

    vals = flex.grid.at_node["lithosphere_surface__increment_of_elevation"]
    assert_array_equal(vals, 0.)

    assert np.may_share_memory(vals, flex.dz_at_node)
    assert flex.dz_at_node.shape == (3, 5)


def test_load_at_node():
    """Test load_at_node is reshaped and shares memory with its field."""
    flex = Flexure1D(RasterModelGrid((3, 5)))

    vals = flex.grid.at_node["lithosphere__increment_of_overlying_pressure"]
    assert_array_equal(vals, 0.)

    assert np.may_share_memory(vals, flex.load_at_node)
    assert flex.load_at_node.shape == (3, 5)


def test_x_is_contiguous():
    """Test that x_at_node is contiguous."""
    flex = Flexure1D(RasterModelGrid((3, 5)))
    assert flex.x_at_node.flags["C_CONTIGUOUS"]


def test_dz_is_contiguous():
    """Test that dz_at_node is contiguous."""
    flex = Flexure1D(RasterModelGrid((3, 5)))
    assert flex.dz_at_node.flags["C_CONTIGUOUS"]


def test_load_is_contiguous():
    """Test that load_at_node is contiguous."""
    flex = Flexure1D(RasterModelGrid((3, 5)))
    assert flex.load_at_node.flags["C_CONTIGUOUS"]



#!/usr/env/python

"""
flow_accumulator.py: Component to accumulate flow and calculate drainage area.

Provides the FlowAccumulator component which accumulates flow and calculates
drainage area. FlowAccumulator supports multiple methods for calculating flow
direction. Optionally a depression finding component can be specified and flow
directing, depression finding, and flow routing can all be accomplished
together.
"""

from __future__ import print_function

import warnings

from landlab import FieldError, Component
from landlab import RasterModelGrid, VoronoiDelaunayGrid  # for type tests
from landlab.utils.return_array import return_array_at_node
from landlab.core.messages import warning_message

from landlab.components.flow_accum import flow_accum_bw
from landlab.components.flow_accum import flow_accum_to_n

from landlab import BAD_INDEX_VALUE
import six
import numpy as np


class FlowAccumulator(Component):

    """
    Component to accumulate flow and calculate drainage area.

    This is accomplished by first finding flow directions by a user-specified
    method and then calculating the drainage area and discharge.

    Optionally, spatially variable runoff can be set either by the model grid
    field 'water__unit_flux_in' or the input variable *runoff_rate**.

    Optionally a depression finding component can be specified and flow
    directing, depression finding, and flow routing can all be accomplished
    together.


    NOTE: The perimeter nodes  NEVER contribute to the accumulating flux, even
    if the  gradients from them point inwards to the main body of the grid.
    This is because under Landlab definitions, perimeter nodes lack cells, so
    cannot accumulate any discharge.

    FlowAccumulator stores as ModelGrid fields:

        -  Node array of drainage areas: *'drainage_area'*
        -  Node array of discharges: *'surface_water__discharge'*
        -  Node array containing downstream-to-upstream ordered list of node
           IDs: *'flow__upstream_node_order'*
        -  Node array of all but the first element of the delta data structure:
            *flow__data_structure_delta*. The first element is always zero.
        -  Link array of the D data structure: *flow__data_structure_D*

    The FlowDirector component will add additional ModelGrid fields.
    DirectToOne methods(Steepest/D4 and D8) and DirectToMany(NAMES HERE) use
    different model grid fields.

    DirectToOne Methods (Steeptest/D4 and D8) store the following as ModelGrid
    fields:

        -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
           there is no receiver: *'flow__receiver_node'*
        -  Node array of steepest downhill slopes:
           *'topographic__steepest_slope'*
        -  Node array containing ID of link that leads from each node to its
           receiver, or BAD_INDEX_VALUE if no link:
           *'flow__link_to_receiver_node'*
        -  Boolean node array of all local lows: *'flow__sink_flag'*

    DirectToMany Methods (MFD) store the following as ModelGrid
    fields:

        -  Node array of receivers (nodes that receive flow), or ITS OWN ID if
           there is no receiver: *'flow__receiver_nodes'*. This array is 2D, and is
           of dimension (number of nodes x max number of receivers).
        -  Node array of flow proportions: *'flow__receiver_proportions'*. This
           array is 2D, and is of dimension (number of nodes x max number of
           receivers).
        -  Node array of links carrying flow:  *'flow__links_to_receiver_nodes'*.
           This array is 2D, and is of dimension (number of nodes x max number of
           receivers).
        -  Node array of the steepest downhill receiver. *'flow__receiver_nodes'*
        -  Node array of steepest downhill slope from each receiver:
           *'topographic__steepest_slope'*
        -  Node array containing ID of steepest link that leads from each node to a
           receiver, or BAD_INDEX_VALUE if no link:
           *'flow__link_to_receiver_node'*
        -  Boolean node array of all local lows: *'flow__sink_flag'*

    The primary method of this class is :func:`run_one_step`

    `run_one_step` takes the optional argument update_flow_director (default is
    True) that determines if the flow_director is re-run before flow is
    accumulated.

    Parameters
    ----------
    grid : ModelGrid
        A grid of type Voroni.
    surface : field name at node or array of length node
        The surface to direct flow across.
    flow_director : string, class, instance of class.
        A string of method or class name (e.g. 'D8' or 'FlowDirectorD8'), an
        uninstantiated FlowDirector class, or an instance of a FlowDirector
        class. This sets the method used to calculate flow directions.
        Default is 'FlowDirectorSteepest'
    runoff_rate : field name, array, or float, optional (m/time)
        If provided, sets the runoff rate and will be assigned to the grid field
        'water__unit_flux_in'. If a spatially and and temporally variable runoff
        rate is desired, pass this field name and update the field through model
        run time. If both the field and argument are present at the time of
        initialization, runoff_rate will *overwrite* the field. If neither are
        set, defaults to spatially constant unit input.
    depression_finder : string, class, instance of class, optional
         A string of class name (e.g., 'DepressionFinderAndRouter'), an
         uninstantiated DepressionFinder class, or an instance of a
         DepressionFinder class.
         This sets the method for depression finding.
    **kwargs : any additional parameters to pass to a FlowDirector or
         DepressionFinderAndRouter instance (e.g., partion_method for
         FlowDirectorMFD). This will have no effect if an instantiated component
         is passed using the flow_director or depression_finder keywords.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> mg = RasterModelGrid((3,3), spacing=(1, 1))
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> _ = mg.add_field('topographic__elevation',
    ...                  mg.node_x + mg.node_y,
    ...                  at = 'node')

    The FlowAccumulator component accumulates flow and calculates drainage using
    all of the different methods for directing flow in Landlab. These include
    steepest descent (also known as D4 for the case of a raster grid) and D8 (on
    raster grids only). The method for flow director can be specified as a
    string (e.g., 'D8' or 'FlowDirectorD8'), as an uninstantiated FlowDirector
    component or as an instantiated FlowDirector component.

    The default method is to use FlowDirectorSteepest.

    First let's look at the three ways to instantiate a FlowAccumulator. The
    following four methods are all equivalent. First, we can pass the entire
    name of a flow director as a string to the argument `flow_director`:

    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                      flow_director='FlowDirectorSteepest')

    Second, we can pass just the method name as a string to the argument
    `flow_director`:

    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                      flow_director='Steepest')

    Third, we can import a FlowDirector component from Landlab and pass it to
    `flow_director`:
    >>> from landlab.components import FlowDirectorSteepest
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                      flow_director=FlowDirectorSteepest)

    Finally, we can instantiate a FlowDirector component and pass this
    instantiated version to `flow_director`. You might want to do this if you
    used a FlowDirector in order to set up something before starting a
    time loop and then want to use the same flow director within the loop.

    >>> fd = FlowDirectorSteepest(mg, 'topographic__elevation')
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                      flow_director=FlowDirectorSteepest)

    Now let's look at what FlowAccumulator does. Even before we run
    FlowAccumulator it has the property `surface_values` that stores the values
    of the surface over which flow is directed and accumulated.

    >>> fa.surface_values
    array([ 0.,  1.,  2.,  1.,  2.,  3.,  2.,  3.,  4.])

    Now let's make a more complicated elevation grid for the next examples.

    >>> mg = RasterModelGrid((5, 4), spacing=(1, 1))
    >>> topographic__elevation = np.array([0.,  0.,  0., 0.,
    ...                                    0., 21., 10., 0.,
    ...                                    0., 31., 20., 0.,
    ...                                    0., 32., 30., 0.,
    ...                                    0.,  0.,  0., 0.])
    >>> _ = mg.add_field('node',
    ...                    'topographic__elevation',
    ...                    topographic__elevation)
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                        flow_director=FlowDirectorSteepest)
    >>> fa.run_one_step()
    >>> mg.at_node['flow__receiver_node'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0,  1,  2,  3,
            4,  1,  2,  7,
            8, 10,  6, 11,
           12, 14, 10, 15,
           16, 17, 18, 19])
    >>> mg.at_node['drainage_area'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0.,  1.,  5.,  0.,
            0.,  1.,  5.,  0.,
            0.,  1.,  4.,  0.,
            0.,  1.,  2.,  0.,
            0.,  0.,  0.,  0.])

    Now let's change the cell area (100.) and the runoff rates:

    >>> mg = RasterModelGrid((5, 4), spacing=(10., 10))

    Put the data back into the new grid.

    >>> _ = mg.add_field('node',
    ...                    'topographic__elevation',
    ...                    topographic__elevation)
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                        flow_director=FlowDirectorSteepest)
    >>> runoff_rate = np.arange(mg.number_of_nodes)
    >>> _ = mg.add_field('node', 'water__unit_flux_in', runoff_rate,
    ...                  noclobber=False)
    >>> fa.run_one_step()
    >>> mg.at_node['surface_water__discharge'] # doctest: +NORMALIZE_WHITESPACE
    array([    0.,   500.,  5200.,     0.,
               0.,   500.,  5200.,     0.,
               0.,   900.,  4600.,     0.,
               0.,  1300.,  2700.,     0.,
               0.,     0.,     0.,     0.])

    The FlowAccumulator component will work for both raster grids and irregular
    grids. For the example we will use a Hexagonal Model Grid, a special type of
    Voroni Grid that has regularly spaced hexagonal cells.

    >>> from landlab import HexModelGrid
    >>> hmg = HexModelGrid(5,3)
    >>> _ = hmg.add_field('topographic__elevation',
    ...                   hmg.node_x + np.round(hmg.node_y),
    ...                   at = 'node')
    >>> fa = FlowAccumulator(hmg, 'topographic__elevation',
    ...                      flow_director=FlowDirectorSteepest)
    >>> fa.surface_values
    array([ 0. ,  1. ,  2. ,
            0.5,  1.5,  2.5,  3.5,
            1. ,  2. ,  3. ,  4. ,  5. ,
            2.5,  3.5,  4.5,  5.5,
            3. ,  4. ,  5. ])

    If the FlowDirector you want to use takes keyword arguments and you want
    to specify it using a string or uninstantiated FlowDirector class, include
    those keyword arguments when you create FlowAccumulator.

    For example, in the case of a raster grid, FlowDirectorMFD can use only
    orthogonal links, or it can use both orthogonal and diagonal links.

    >>> mg = RasterModelGrid((5, 5), spacing=(1, 1))
    >>> topographic__elevation = mg.node_y+mg.node_x
    >>> _ = mg.add_field('node',
    ...                  'topographic__elevation',
    ...                   topographic__elevation)
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                      flow_director='MFD',
    ...                      diagonals = True)
    >>> fa.run_one_step()
    >>> mg.at_node['flow__receiver_nodes'] # doctest: +NORMALIZE_WHITESPACE
    array([[ 0, -1, -1, -1, -1, -1, -1, -1],
           [ 1, -1, -1, -1, -1, -1, -1, -1],
           [ 2, -1, -1, -1, -1, -1, -1, -1],
           [ 3, -1, -1, -1, -1, -1, -1, -1],
           [ 4, -1, -1, -1, -1, -1, -1, -1],
           [ 5, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1,  5,  1, -1, -1,  0, -1],
           [-1, -1,  6,  2, -1, -1,  1, -1],
           [-1, -1,  7,  3, -1, -1,  2, -1],
           [ 9, -1, -1, -1, -1, -1, -1, -1],
           [10, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, 10,  6, -1, -1,  5, -1],
           [-1, -1, 11,  7, -1, -1,  6, -1],
           [-1, -1, 12,  8, -1, -1,  7, -1],
           [14, -1, -1, -1, -1, -1, -1, -1],
           [15, -1, -1, -1, -1, -1, -1, -1],
           [-1, -1, 15, 11, -1, -1, 10, -1],
           [-1, -1, 16, 12, -1, -1, 11, -1],
           [-1, -1, 17, 13, -1, -1, 12, -1],
           [19, -1, -1, -1, -1, -1, -1, -1],
           [20, -1, -1, -1, -1, -1, -1, -1],
           [21, -1, -1, -1, -1, -1, -1, -1],
           [22, -1, -1, -1, -1, -1, -1, -1],
           [23, -1, -1, -1, -1, -1, -1, -1],
           [24, -1, -1, -1, -1, -1, -1, -1]])
    >>> mg.at_node['drainage_area'] # doctest: +NORMALIZE_WHITESPACE
    array([ 1.41168825,  2.06497116,  1.3253788 ,  0.40380592,  0.        ,
            2.06497116,  3.40811691,  2.5753788 ,  1.37867966,  0.        ,
            1.3253788 ,  2.5753788 ,  2.17157288,  1.29289322,  0.        ,
            0.40380592,  1.37867966,  1.29289322,  1.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.        ,  0.        ])

    It may seem odd that there are no round numbers in the drainage area field.
    This is because flow is directed to all downhill boundary nodes and
    partitioned based on slope.

    To check that flow is conserved, sum along all boundary nodes.

    >>> sum(mg.at_node['drainage_area'][mg.boundary_nodes])
    9.0000000000000018

    This should be the same as the number of core nodes --- as boundary nodes
    in landlab do not have area.

    >>> len(mg.core_nodes)
    9

    Next, let's set the dx spacing such that each cell has an area of one.

    >>> dx=(2./(3.**0.5))**0.5
    >>> hmg = HexModelGrid(5,3, dx)
    >>> _ = hmg.add_field('topographic__elevation',
    ...                     hmg.node_x**2 + np.round(hmg.node_y)**2,
    ...                     at = 'node')
    >>> fa = FlowAccumulator(hmg, 'topographic__elevation',
    ...                        flow_director=FlowDirectorSteepest)
    >>> fa.run_one_step()
    >>> hmg.at_node['flow__receiver_node'] # doctest: +NORMALIZE_WHITESPACE
    array([ 0,  1,  2,
            3,  0,  1,  6,
            7,  3,  4,  5, 11,
           12,  8,  9, 15,
           16, 17, 18])
    >>> hmg.at_node['drainage_area'] # doctest: +NORMALIZE_WHITESPACE
    array([ 3.,  2.,  0.,
            2.,  3.,  2.,  0.,
            0.,  2.,  2.,  1.,  0.,
            0., 1.,  1.,  0.,
            0.,  0.,  0.])

    Now let's change the cell area (100.) and the runoff rates:

    >>> hmg = HexModelGrid(5,3, dx*10.)

    Put the data back into the new grid.

    >>> _ = hmg.add_field('topographic__elevation',
    ...                     hmg.node_x**2 + np.round(hmg.node_y)**2,
    ...                     at = 'node')
    >>> fa = FlowAccumulator(hmg, 'topographic__elevation',
    ...                        flow_director=FlowDirectorSteepest)
    >>> fa.run_one_step()
    >>> hmg.at_node['surface_water__discharge']
    array([ 500.,    0.,    0.,
            200.,  500.,  200.,    0.,
              0.,  200.,  200.,  100.,    0.,
              0.,  100.,  100.,    0.,
              0.,    0.,    0.])

    Next, let's see what happens to a raster grid when there is a depression.

    >>> mg = RasterModelGrid((7, 7), 0.5)
    >>> z = mg.add_field('node', 'topographic__elevation', mg.node_x.copy())
    >>> z += 0.01 * mg.node_y
    >>> mg.at_node['topographic__elevation'].reshape(mg.shape)[2:5, 2:5] *= 0.1
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, False, True)

    This model grid has a depression in the center.

    >>> mg.at_node['topographic__elevation'].reshape(mg.shape)
    array([[ 0.    ,  0.5   ,  1.    ,  1.5   ,  2.    ,  2.5   ,  3.    ],
           [ 0.005 ,  0.505 ,  1.005 ,  1.505 ,  2.005 ,  2.505 ,  3.005 ],
           [ 0.01  ,  0.51  ,  0.101 ,  0.151 ,  0.201 ,  2.51  ,  3.01  ],
           [ 0.015 ,  0.515 ,  0.1015,  0.1515,  0.2015,  2.515 ,  3.015 ],
           [ 0.02  ,  0.52  ,  0.102 ,  0.152 ,  0.202 ,  2.52  ,  3.02  ],
           [ 0.025 ,  0.525 ,  1.025 ,  1.525 ,  2.025 ,  2.525 ,  3.025 ],
           [ 0.03  ,  0.53  ,  1.03  ,  1.53  ,  2.03  ,  2.53  ,  3.03  ]])
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                        flow_director=FlowDirectorSteepest)
    >>> fa.run_one_step()  # the flow "gets stuck" in the hole
    >>> mg.at_node['flow__receiver_node'].reshape(mg.shape)
    array([[ 0,  1,  2,  3,  4,  5,  6],
           [ 7,  7, 16, 17, 18, 11, 13],
           [14, 14, 16, 16, 17, 18, 20],
           [21, 21, 16, 23, 24, 25, 27],
           [28, 28, 23, 30, 31, 32, 34],
           [35, 35, 30, 31, 32, 39, 41],
           [42, 43, 44, 45, 46, 47, 48]])
    >>> mg.at_node['drainage_area'].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 0.25,  0.25,  0.25,  0.25,  0.5 ,  0.25,  0.  ],
           [ 0.25,  0.25,  5.  ,  1.5 ,  1.  ,  0.25,  0.  ],
           [ 0.25,  0.25,  3.  ,  0.75,  0.5 ,  0.25,  0.  ],
           [ 0.25,  0.25,  2.  ,  1.5 ,  1.  ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.25,  0.25,  0.5 ,  0.25,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])

    Because of the depression, the flow 'got stuck' in the hole in the center
    of the grid. We can fix this by using a depression finder, such as
    DepressionFinderAndRouter.

    >>> from landlab.components import DepressionFinderAndRouter

    We can either run the depression finder separately from the flow
    accumulator or we can specify the depression finder and router when we
    instantiate the accumulator and it will run automatically. Similar to
    specifying the FlowDirector we can provide a depression finder in multiple
    three ways.

    First let's try running them separately.

    >>> df_4 = DepressionFinderAndRouter(mg)
    >>> df_4.map_depressions()
    >>> mg.at_node['flow__receiver_node'].reshape(mg.shape)
    array([[ 0,  1,  2,  3,  4,  5,  6],
           [ 7,  7, 16, 17, 18, 11, 13],
           [14, 14,  8, 16, 17, 18, 20],
           [21, 21, 16, 16, 24, 25, 27],
           [28, 28, 23, 24, 24, 32, 34],
           [35, 35, 30, 31, 32, 39, 41],
           [42, 43, 44, 45, 46, 47, 48]])
    >>> mg.at_node['drainage_area'].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 5.25,  5.25,  0.25,  0.25,  0.5 ,  0.25,  0.  ],
           [ 0.25,  0.25,  5.  ,  1.5 ,  1.  ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.75,  2.25,  0.5 ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.5 ,  0.5 ,  1.  ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.25,  0.25,  0.5 ,  0.25,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])

    Now the flow is routed correctly. The depression finder has properties that
    including whether there is a lake at the node, which lake is at each node,
    the outlet node of each lake, and the area of each lake.

    >>> df_4.lake_at_node.reshape(mg.shape)  # doctest: +NORMALIZE_WHITESPACE
    array([[False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False],
           [False, False,  True,  True,  True, False, False],
           [False, False,  True,  True,  True, False, False],
           [False, False,  True,  True,  True, False, False],
           [False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False]], dtype=bool)
    >>> df_4.lake_map.reshape(mg.shape)  # doctest: +NORMALIZE_WHITESPACE
    array([[-1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1],
           [-1, -1, 16, 16, 16, -1, -1],
           [-1, -1, 16, 16, 16, -1, -1],
           [-1, -1, 16, 16, 16, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1]])
    >>> df_4.lake_codes  # a unique code for each lake present on the grid
    array([16])
    >>> df_4.lake_outlets  # the outlet node of each lake in lake_codes
    array([8])
    >>> df_4.lake_areas  # the area of each lake in lake_codes
    array([ 2.25])

    Alternatively, we can initialize a flow accumulator with a depression
    finder specified. Calling run_one_step() will run both the accumulator
    and the depression finder with one call. For this example, we will pass the
    class DepressionFinderAndRouter to the parameter `depression_finder`.

    >>> mg = RasterModelGrid((7, 7), 0.5)
    >>> z = mg.add_field('node', 'topographic__elevation', mg.node_x.copy())
    >>> z += 0.01 * mg.node_y
    >>> mg.at_node['topographic__elevation'].reshape(mg.shape)[2:5, 2:5] *= 0.1
    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                        flow_director='FlowDirectorD8',
    ...                        depression_finder=DepressionFinderAndRouter)
    >>> fa.run_one_step()

    This has the same effect of first calling the accumulator and then calling
    the depression finder.

    >>> mg.at_node['flow__receiver_node'].reshape(mg.shape)
    array([[ 0,  1,  2,  3,  4,  5,  6],
           [ 7,  7, 16, 17, 18, 18, 13],
           [14, 14,  8, 16, 17, 18, 20],
           [21, 21, 16, 16, 24, 25, 27],
           [28, 28, 23, 24, 24, 32, 34],
           [35, 35, 30, 31, 32, 32, 41],
           [42, 43, 44, 45, 46, 47, 48]])
    >>> mg.at_node['drainage_area'].reshape(mg.shape)
    array([[ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
           [ 5.25,  5.25,  0.25,  0.25,  0.25,  0.25,  0.  ],
           [ 0.25,  0.25,  5.  ,  1.5 ,  1.  ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.75,  2.25,  0.5 ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.5 ,  0.5 ,  1.  ,  0.25,  0.  ],
           [ 0.25,  0.25,  0.25,  0.25,  0.25,  0.25,  0.  ],
           [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ,  0.  ]])

    The depression finder is stored as part of the flow accumulator, so its
    properties can be accessed through the depression finder.

    >>> fa.depression_finder.lake_at_node.reshape(mg.shape)  # doctest: +NORMALIZE_WHITESPACE
    array([[False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False],
           [False, False,  True,  True,  True, False, False],
           [False, False,  True,  True,  True, False, False],
           [False, False,  True,  True,  True, False, False],
           [False, False, False, False, False, False, False],
           [False, False, False, False, False, False, False]], dtype=bool)
    >>> fa.depression_finder.lake_map.reshape(mg.shape)  # doctest: +NORMALIZE_WHITESPACE
    array([[-1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1],
           [-1, -1, 16, 16, 16, -1, -1],
           [-1, -1, 16, 16, 16, -1, -1],
           [-1, -1, 16, 16, 16, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1],
           [-1, -1, -1, -1, -1, -1, -1]])
    >>> fa.depression_finder.lake_codes  # a unique code for each lake present on the grid
    array([16])
    >>> fa.depression_finder.lake_outlets  # the outlet node of each lake in lake_codes
    array([8])
    >>> fa.depression_finder.lake_areas  # the area of each lake in lake_codes
    array([ 2.25])

    Finally, note that the DepressionFinderAndRouter takes a keyword argument
    *routing* ('D8', default; 'D4') that sets how connectivity is set between
    nodes. Similar to our ability to pass keyword arguments to the FlowDirector
    through FlowAccumulator, we can pass this keyword argument to the
    DepressionFinderAndRouter component.

    >>> fa = FlowAccumulator(mg, 'topographic__elevation',
    ...                      flow_director=FlowDirectorSteepest,
    ...                      depression_finder=DepressionFinderAndRouter,
    ...                      routing='D4')

    """

    _name = 'FlowAccumulator'

    _input_var_names = ('topographic__elevation',
                        'water__unit_flux_in'
                        )

    _output_var_names = ('drainage_area',
                         'surface_water__discharge',
                         'flow__upstream_node_order',
                         'flow__nodes_not_in_stack',
                         'flow__data_structure_delta',
                         'flow__data_structure_D'
                         )

    _var_units = {'topographic__elevation': 'm',
                  'flow__receiver_node': 'm',
                  'water__unit_flux_in': 'm/s',
                  'drainage_area': 'm**2',
                  'surface_water__discharge': 'm**3/s',
                  'flow__upstream_node_order': '-',
                  'flow__data_structure_delta': '-',
                  'flow__data_structure_D': '-',
                  'flow__nodes_not_in_stack': '-'
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'flow__receiver_node': 'node',
                    'water__unit_flux_in': 'node',
                    'drainage_area': 'node',
                    'surface_water__discharge': 'node',
                    'flow__upstream_node_order': 'node',
                    'flow__nodes_not_in_stack': 'grid',
                    'flow__data_structure_delta': 'node',
                    'flow__data_structure_D': 'link',
                    }
    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
        'drainage_area':
            "Upstream accumulated surface area contributing to the node's "
            "discharge",
        'surface_water__discharge': 'Discharge of water through each node',
        'water__unit_flux_in':
            'External volume water per area per time input to each node '
            '(e.g., rainfall rate)',
        'flow__upstream_node_order':
            'Node array containing downstream-to-upstream ordered list of '
            'node IDs',
        'flow__data_structure_delta':
            'Node array containing the elements delta[1:] of the data '
            'structure "delta" used for construction of the downstream-to-'
            'upstream node array',
        'flow__data_structure_D':
            'Link array containing the data structure D used for construction'
            'of the downstream-to-upstream node array',
        'flow__nodes_not_in_stack':
            'Boolean value indicating if there are any nodes that have not yet'
            'been added to the stack stored in flow__upstream_node_order.'
               }

    def __init__(self,
                 grid,
                 surface='topographic__elevation',
                 flow_director='FlowDirectorSteepest',
                 runoff_rate=None,
                 depression_finder=None,
                 **kwargs):
        """
        Initialize the FlowAccumulator component.

        Saves the grid, tests grid type, tests imput types and compatability
        for the flow_director and depression_finder keyword arguments, tests
        the argument of runoff_rate, and initializes new fields.
        """
        super(FlowAccumulator, self).__init__(grid)
        # Keep a local reference to the grid
        self._grid = grid

        # Grid type testing
        self._is_raster = isinstance(self._grid, RasterModelGrid)
        self._is_Voroni = isinstance(self._grid, VoronoiDelaunayGrid)

        self.kwargs = kwargs
        # STEP 1: Testing of input values, supplied either in function call or
        # as part of the grid.
        self._test_water_inputs(grid, runoff_rate)

        # save elevations and node_cell_area to class properites.
        self.surface = surface
        self.surface_values = return_array_at_node(grid, surface)

        node_cell_area = self._grid.cell_area_at_node.copy()
        node_cell_area[self._grid.closed_boundary_nodes] = 0.

        self.node_cell_area = node_cell_area

        # STEP 2:
        # identify Flow Director method, save name, import and initialize the correct
        # flow director component if necessary
        self._add_director(flow_director)
        self._add_depression_finder(depression_finder)

        # This component will track of the following variables.
        # Attempt to create each, if they already exist, assign the existing
        # version to the local copy.

        #   - drainage area at each node
        #   - receiver of each node
        #   - D array
        #   - delta array
        #   - missing nodes in stack.
        try:
            self.drainage_area = grid.add_zeros('drainage_area', at='node',
                                                dtype=float)
        except FieldError:
            self.drainage_area = grid.at_node['drainage_area']

        try:
            self.discharges = grid.add_zeros('surface_water__discharge',
                                             at='node', dtype=float)
        except FieldError:
            self.discharges = grid.at_node['surface_water__discharge']

        try:
            self.upstream_ordered_nodes = grid.add_field('flow__upstream_node_order',
                                                         BAD_INDEX_VALUE*grid.ones(at='node', dtype=int),
                                                         at='node', dtype=int)

        except FieldError:
            self.upstream_ordered_nodes = grid.at_node[
                'flow__upstream_node_order']

        try:
            self.delta_structure = grid.add_field('flow__data_structure_delta',
                                                  BAD_INDEX_VALUE*grid.ones(at='node', dtype=int),
                                                  at='node', dtype=int)
        except FieldError:
            self.delta_structure = grid.at_node['flow__data_structure_delta']

        try:

            if self.flow_director.to_n_receivers == 'many' and self._is_raster:
                # needs to be BAD_INDEX_VALUE
                self.D_structure = grid.add_field('flow__data_structure_D',
                                                  BAD_INDEX_VALUE*np.ones((self._grid.number_of_links, 2),
                                                  dtype=int),
                                                  at='link',
                                                  dtype=int,
                                                  noclobber=False)
            else:

                # needs to be BAD_INDEX_VALUE
                self.D_structure = grid.add_field('flow__data_structure_D',
                                                  BAD_INDEX_VALUE*grid.ones(at='link'),
                                                  at='link', dtype=int)
        except FieldError:
            self.D_structure = grid.at_link['flow__data_structure_D']

        self.nodes_not_in_stack = True

    @property
    def node_drainage_area(self):
        """Return the drainage area."""
        return self._grid['node']['drainage_area']

    @property
    def node_water_discharge(self):
        """Return the surface water discharge."""
        return self._grid['node']['surface_water__discharge']

    @property
    def node_order_upstream(self):
        """Return the upstream node order (drainage stack)."""
        return self._grid['node']['flow__upstream_node_order']

    def _test_water_inputs(self, grid, runoff_rate):
        """Test inputs for runoff_rate and water__unit_flux_in."""
        if 'water__unit_flux_in' not in grid.at_node:
            if runoff_rate is None:
                # assume that if runoff rate is not supplied, that the value
                # should be set to one everywhere.
                grid.add_ones('node', 'water__unit_flux_in', dtype=float)
            else:
                runoff_rate = return_array_at_node(grid, runoff_rate)
                grid.at_node['water__unit_flux_in'] = runoff_rate
        else:
            if runoff_rate is not None:
                print ("FlowAccumulator found both the field " +
                       "'water__unit_flux_in' and a provided float or " +
                       "array for the runoff_rate argument. THE FIELD IS " +
                       "BEING OVERWRITTEN WITH THE SUPPLIED RUNOFF_RATE!")
                runoff_rate = return_array_at_node(grid, runoff_rate)
                grid.at_node['water__unit_flux_in'] = runoff_rate

        # perform a test (for politeness!) that the old name for the water_in
        # field is not present:
        if 'water__discharge_in' in grid.at_node:
            warnings.warn("This component formerly took 'water__discharge" +
                          "_in' as an input field. However, this field is " +
                          "now named 'water__unit_flux_in'. You are still " +
                          "using a field with the old name. Please update " +
                          "your code if you intended the FlowRouter to use " +
                          "that field.", DeprecationWarning)

    def _add_director(self, flow_director):
        """Test and add the flow director component."""
        PERMITTED_DIRECTORS = ['FlowDirectorSteepest',
                               'FlowDirectorD8',
                               'FlowDirectorMFD',
                               'FlowDirectorDINF']

        # find keyword args to pass along:
        try:
            component_name = flow_director._name
        except:
            component_name = None

        potential_kwargs = ['partition_method', 'diagonals']
        kw = {}
        for p_k in potential_kwargs:
            if p_k in self.kwargs.keys():
                kw[p_k] = self.kwargs.pop(p_k)

        # flow director is provided as a string.
        if isinstance(flow_director, six.string_types):
            if flow_director[:12] == 'FlowDirector':
                flow_director = flow_director[12:]

            from landlab.components.flow_director import (FlowDirectorSteepest,
                                                          FlowDirectorD8,
                                                          FlowDirectorMFD,
                                                          FlowDirectorDINF)
            DIRECTOR_METHODS = {'D4': FlowDirectorSteepest,
                                'Steepest': FlowDirectorSteepest,
                                'D8': FlowDirectorD8,
                                'MFD': FlowDirectorMFD,
                                'DINF': FlowDirectorDINF
                                }

            try:
                FlowDirector = DIRECTOR_METHODS[flow_director]
            except KeyError:
                raise ValueError('String provided in flow_director is not a '
                                 'valid method or component name. The following'
                                 'components are valid imputs:\n'\
                                 + str(PERMITTED_DIRECTORS))
            self.flow_director = FlowDirector(self._grid, self.surface, **kw)
        # flow director is provided as an instantiated flow director
        elif isinstance(flow_director, Component):
            if flow_director._name in PERMITTED_DIRECTORS:
                self.flow_director = flow_director
            else:
                raise ValueError('String provided in flow_director is not a '
                                 'valid method or component name. The following'
                                 'components are valid imputs:\n'\
                                 + str(PERMITTED_DIRECTORS))

            if len(kw) > 0:
                raise ValueError('flow_director provided as an instantiated ',
                                 'component and keyword arguments provided. ',
                                 'These kwargs would be ignored.')

        # flow director is provided as an uninstantiated flow director
        else:
            if flow_director._name in PERMITTED_DIRECTORS:
                FlowDirector = flow_director
                self.flow_director = FlowDirector(self._grid, self.surface, **kw)
            else:
                raise ValueError('String provided in flow_director is not a '
                                 'valid method or component name. The following'
                                 'components are valid imputs:\n'\
                                 + str(PERMITTED_DIRECTORS))

        # save method as attribute
        self.method = self.flow_director.method

    def _add_depression_finder(self, depression_finder):
        """Test and add the depression finder component."""
        PERMITTED_DEPRESSION_FINDERS = ['DepressionFinderAndRouter']

        # now do a similar thing for the depression finder.
        self.depression_finder_provided = depression_finder
        if self.depression_finder_provided is not None:

            # collect potential kwargs to pass to depression_finder
            # instantiation
            potential_kwargs = ['routing']
            kw = {}
            for p_k in potential_kwargs:
                if p_k in self.kwargs.keys():
                    kw[p_k] = self.kwargs.pop(p_k)

            # NEED TO TEST WHICH FLOWDIRECTOR WAS PROVIDED.
            if self.flow_director._name in ('FlowDirectorMFD',
                                            'FlowDirectorDINF'):
                raise ValueError('The depression finder only works with route '
                                 'to one FlowDirectors such as '
                                 'FlowDirectorSteepest and  FlowDirectorD8. '
                                 'Provide a different FlowDirector.')

            # if D4 is being used here and should be.
            if ((('routing' not in kw) or (kw['routing'] != 'D4')) and
                isinstance(self._grid, RasterModelGrid) and
                (self.flow_director._name in('FlowDirectorSteepest'))):

                message = ('You have specified \n'
                           'flow_director=FlowDirectorSteepest and\n'
                           'depression_finder=DepressionFinderAndRouter\n'
                           'in the instantiation of FlowAccumulator on a '
                           'RasterModelGrid. The default behavior of '
                           'DepressionFinderAndRouter is to use D8 connectivity '
                           'which is in conflict with D4 connectivity used by '
                           'FlowDirectorSteepest. \n'
                           "To fix this, provide the kwarg routing='D4', when "
                           'you instantiate FlowAccumulator.')

                raise ValueError(warning_message(message))

            # depression finder is provided as a string.
            if isinstance(self.depression_finder_provided, six.string_types):

                from landlab.components import DepressionFinderAndRouter
                DEPRESSION_METHODS = {'DepressionFinderAndRouter': DepressionFinderAndRouter
                                    }

                try:
                    DepressionFinder = DEPRESSION_METHODS[self.depression_finder_provided]
                except KeyError:
                    raise ValueError('Component provided in depression_finder '
                                     'is not a valid component. The following '
                                     'components are valid imputs:\n' \
                                     + str(PERMITTED_DEPRESSION_FINDERS))

                self.depression_finder = DepressionFinder(self._grid, **kw)
            # flow director is provided as an instantiated depression finder
            elif isinstance(self.depression_finder_provided, Component):

                if self.depression_finder_provided._name in PERMITTED_DEPRESSION_FINDERS:
                    self.depression_finder = self.depression_finder_provided
                else:
                    raise ValueError('Component provided in depression_finder '
                                     'is not a valid component. The following '
                                     'components are valid imputs:\n' \
                                     + str(PERMITTED_DEPRESSION_FINDERS))

                if len(kw) > 0:
                    raise ValueError('flow_director provided as an instantiated ',
                                     'component and keyword arguments provided. ',
                                     'These kwargs would be ignored.')

                # if D4 is being used here and should be.
                if (self.depression_finder._D8 and
                    (self.flow_director._name in ('FlowDirectorSteepest'))):

                    message = ('You have specified \n'
                               'flow_director=FlowDirectorSteepest and\n'
                               'depression_finder=DepressionFinderAndRouter\n'
                               'in the instantiation of FlowAccumulator on a '
                               'RasterModelGrid. The behavior of the instantiated '
                               'DepressionFinderAndRouter is to use D8 connectivity '
                               'which is in conflict with D4 connectivity used by '
                               'FlowDirectorSteepest. \n'
                               "To fix this, provide the kwarg routing='D4', when "
                               'you instantiate DepressionFinderAndRouter.')

                    raise ValueError(warning_message(message))

            # depression_fiuner is provided as an uninstantiated depression finder
            else:

                if self.depression_finder_provided._name in PERMITTED_DEPRESSION_FINDERS:
                    DepressionFinder = self.depression_finder_provided
                    self.depression_finder = DepressionFinder(self._grid, **kw)
                else:
                    raise ValueError('Component provided in depression_finder '
                                     'is not a valid component. The following '
                                     'components are valid imputs:\n' \
                                     + str(PERMITTED_DEPRESSION_FINDERS))
        else:
            self.depression_finder = None

    def accumulate_flow(self, update_flow_director=True):
        """
        Function to make FlowAccumulator calculate drainage area and discharge.

        Running run_one_step() results in the following to occur:
            1. Flow directions are updated (unless update_flow_director is set
            as False).
            2. Intermediate steps that analyse the drainage network topology
            and create datastructures for efficient drainage area and discharge
            calculations.
            3. Calculation of drainage area and discharge.
            4. Depression finding and mapping, which updates drainage area and
            discharge.
        """
        # step 1. Find flow directions by specified method
        if update_flow_director == True:
            self.flow_director.run_one_step()

        # further steps vary depending on how many recievers are present
        # one set of steps is for route to one (D8, Steepest/D4)
        if self.flow_director.to_n_receivers == 'one':

            # step 3. Run depression finder if passed
            # Depression finder reaccumulates flow at the end of its routine.
            if self.depression_finder_provided is not None:

                self.depression_finder.map_depressions()

                a = self._grid['node']['drainage_area']
                q = self._grid['node']['surface_water__discharge']

            else:
                # step 2. Get r
                r = self._grid['node']['flow__receiver_node']

                # step 2. Stack, D, delta construction
                nd = flow_accum_bw._make_number_of_donors_array(r)
                delta = flow_accum_bw._make_delta_array(nd)
                D = flow_accum_bw._make_array_of_donors(r, delta)
                s = flow_accum_bw.make_ordered_node_array(r)

                # put theese in grid so that depression finder can use it.
                # store the generated data in the grid
                self._grid['node']['flow__data_structure_delta'][:] = delta[1:]
                self._grid['link']['flow__data_structure_D'][:len(D)] = D
                self._grid['node']['flow__upstream_node_order'][:] = s

                # step 4. Accumulate (to one or to N depending on direction method. )
                a, q = flow_accum_bw.find_drainage_area_and_discharge(s,
                                                                      r,
                                                                      self.node_cell_area,
                                                                      self._grid.at_node['water__unit_flux_in'])
                self._grid['node']['drainage_area'][:] = a
                self._grid['node']['surface_water__discharge'][:] = q

        else:
            # step 2. Get r and p
            r = self._grid['node']['flow__receiver_nodes']
            p = self._grid['node']['flow__receiver_proportions']

            # step 2. Stack, D, delta construction
            nd = flow_accum_to_n._make_number_of_donors_array_to_n(r, p)
            delta = flow_accum_to_n._make_delta_array_to_n(nd)
            D = flow_accum_to_n._make_array_of_donors_to_n(r, p, delta)
            s = flow_accum_to_n.make_ordered_node_array_to_n(r, p)

            # put theese in grid so that depression finder can use it.
            # store the generated data in the grid
            self._grid['node']['flow__data_structure_delta'][:] = delta[1:]

            if self._is_raster:
                tempD = BAD_INDEX_VALUE * np.ones((self._grid.number_of_links*2))
                tempD[:len(D)] = D
                self._grid['link']['flow__data_structure_D'][:] = tempD.reshape((self._grid.number_of_links, 2))
            else:
                self._grid['link']['flow__data_structure_D'][:len(D)] = D
            self._grid['node']['flow__upstream_node_order'][:] = s

            # step 3. Run depression finder if passed
            # at present this must go at the end.

            # step 4. Accumulate (to one or to N depending on direction method. )
            a, q = flow_accum_to_n.find_drainage_area_and_discharge_to_n(s,
                                                                         r,
                                                                         p,
                                                                         self.node_cell_area,
                                                                         self._grid.at_node['water__unit_flux_in'])
            # store drainage area and discharge.
            self._grid['node']['drainage_area'][:] = a
            self._grid['node']['surface_water__discharge'][:] = q

            # at the moment, this is where the depression finder needs to live.
            if self.depression_finder_provided is not None:
                self.depression_finder.map_depressions()

        return (a, q)

    def run_one_step(self):
        """
        Accumulate flow and save to the model grid.

        run_one_step() checks for updated boundary conditions, calculates
        slopes on links, finds baselevel nodes based on the status at node,
        calculates flow directions, and accumulates flow and saves results to
        the grid.

        An alternative to run_one_step() is accumulate_flow() which does the
        same things but also returns the drainage area and discharge.
        """
        self.accumulate_flow()

if __name__ == '__main__':
    import doctest
    doctest.testmod()



from __future__ import print_function

import os
import sys

import numpy as np


def _deposit_or_erode(layers, n_layers, dz):
    """Add a new layer to the top of a stack.

    Parameters
    ----------
    layers : ndarray of shape `(n_layers, n_nodes)`
        Array of layer thicknesses.
    n_layers : int
        Number of active layers.
    dz : ndarray of shape `(n_nodes, )`
        Thickness of the new layer. Negative thicknesses mean
        erode the top-most layers.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.layers.eventlayers import _deposit_or_erode

    >>> layers = np.full((4, 3), -1.)
    >>> dz = np.array([1., 2., 3.])
    >>> _deposit_or_erode(layers, 1, dz)
    >>> layers
    array([[ 1.,  2.,  3.],
           [-1., -1., -1.],
           [-1., -1., -1.],
           [-1., -1., -1.]])

    >>> dz = np.array([1., 1., 1.])
    >>> _deposit_or_erode(layers, 2, dz)
    >>> layers
    array([[ 1.,  2.,  3.],
           [ 1.,  1.,  1.],
           [-1., -1., -1.],
           [-1., -1., -1.]])

    >>> _deposit_or_erode(layers, 3, [1., -1., -2.])
    >>> layers
    array([[ 1.,  2.,  2.],
           [ 1.,  0.,  0.],
           [ 1.,  0.,  0.],
           [-1., -1., -1.]])
    """
    from .ext.eventlayers import deposit_or_erode

    layers = layers.reshape((layers.shape[0], -1))
    try:
        dz = dz.reshape((layers.shape[1], ))
    except AttributeError:
        dz = np.broadcast_to(dz, (layers.shape[1], ))
    finally:
        dz = np.asfarray(dz)

    deposit_or_erode(layers, n_layers, dz)


def resize_array(array, newsize, exact=False):
    """Increase the size of an array, leaving room to grow.

    Parameters
    ----------
    array : ndarray
        The array to resize.
    newsize : int
        Size of the zero-th dimension of the resized array.
    exact : bool, optional
        Should the new array have the exact size provided or
        at least that size.

    Returns
    -------
    ndarray
        Copy of the input array resized.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.layers.eventlayers import resize_array

    >>> x = np.arange(6)
    >>> bigger_x = resize_array(x, 10)
    >>> bigger_x.size
    17
    >>> np.all(x[:6] == bigger_x[:6])
    True
    >>> x is bigger_x
    False

    >>> x = np.arange(6).reshape((2, 3))
    >>> bigger_x = resize_array(x, 4)
    >>> bigger_x.shape == (10, 3)
    True

    >>> bigger_x = resize_array(x, 4, exact=True)
    >>> bigger_x.shape == (4, 3)
    True
    """
    newsize = int(newsize)
    allocated = array.shape[0]

    if newsize <= allocated:
        return array

    if exact:
        new_allocated = newsize
    else:
        new_allocated = (newsize >> 3) + 6 + newsize

    larger_array = np.empty((new_allocated, ) + array.shape[1:],
                            dtype=array.dtype)
    larger_array[:allocated] = array

    return larger_array


def _allocate_layers_for(array, number_of_layers, number_of_stacks):
    """Allocate a layer matrix.

    Parameters
    ----------
    array : number or ndarray
        Array of layer properties to track.
    number_of_layers : int
        Number of layers to allocate.
    number_of_stacks : int
        Number of stacks to allocate.

    Returns
    -------
    ndarray of size `(number_of_layers, number_of_stacks, values_per_stack)`
        Newly allocated matrix for storing layer properties.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab.layers.eventlayers import _allocate_layers_for

    >>> layers = _allocate_layers_for(3, 2, 4)
    >>> layers.shape == (2, 4)
    True
    >>> layers.dtype.kind == 'i'
    True

    >>> layers = _allocate_layers_for(np.zeros(4), 2, 4)
    >>> layers.shape == (2, 4)
    True
    >>> layers.dtype.kind == 'f'
    True

    >>> layers = _allocate_layers_for(np.zeros(2), 2, 4)
    >>> layers.shape == (2, 4, 2)
    True
    >>> layers.dtype.kind == 'f'
    True
    """
    array = np.asarray(array)

    if array.ndim > 0 and len(array) != number_of_stacks:
        values_per_stack = array.shape
    else:
        values_per_stack = array.shape[1:]
        
    return np.empty((number_of_layers, number_of_stacks) + values_per_stack,
                    dtype=array.dtype)
    

class EventLayersMixIn(object):

    """MixIn that adds a layers attribute to a ModelGrid."""

    @property
    def layers(self):
        """Layers for each cell."""
        try:
            self._layers
        except AttributeError:
            self._layers = EventLayers(self.number_of_cells)
        finally:
            return self._layers


class EventLayers(object):

    """Track layers where each event is its own layer.
    
    Parameters
    ----------
    number_of_stacks : int
        Number of layer stacks to track.

    Examples
    --------
    >>> from landlab.layers.eventlayers import EventLayers

    Create an empty layer stack with 5 stacks.

    >>> layers = EventLayers(5)
    >>> layers.number_of_stacks
    5
    >>> layers.number_of_layers
    0

    Add a layer with a uniform thickness.

    >>> layers.add(1.5)
    >>> layers.number_of_layers
    1
    >>> layers.dz
    array([[ 1.5,  1.5,  1.5,  1.5,  1.5]])

    Add a second layer with uneven thickness.

    >>> layers.add([1., 2., .5, 5., 0.])
    >>> layers.dz
    array([[ 1.5,  1.5,  1.5,  1.5,  1.5],
           [ 1. ,  2. ,  0.5,  5. ,  0. ]])

    Adding a layer with negative thickness will remove
    existing layers for the top of the stack.

    >>> layers.add(-1)
    >>> layers.dz
    array([[ 1.5,  1.5,  1. ,  1.5,  0.5],
           [ 0. ,  1. ,  0. ,  4. ,  0. ],
           [ 0. ,  0. ,  0. ,  0. ,  0. ]])
    """

    def __init__(self, number_of_stacks, allocated=0):
        self._number_of_layers = 0
        self._number_of_stacks = number_of_stacks

        self._attrs = dict()

        dims = (self.number_of_layers, self.number_of_stacks)
        self._attrs['_dz'] = np.empty(dims , dtype=float)
        self._resize(allocated, exact=True)

    def __getitem__(self, name):
        return self._attrs[name][:self.number_of_layers]

    def __str__(self):
        lines = [
            "number_of_layers: {number_of_layers}",
            "number_of_stacks: {number_of_stacks}",
            "tracking: {attrs}",
        ]
        return os.linesep.join(lines).format(
            number_of_layers=self.number_of_layers,
            number_of_stacks=self.number_of_stacks,
            attrs=', '.join(self.tracking) or 'null')

    def __repr__(self):
        return 'EventLayers({number_of_stacks})'.format(
            number_of_stacks=self.number_of_stacks)

    @property
    def tracking(self):
        """Layer properties being tracked.
        
        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers
        >>> layers = EventLayers(3)
        >>> layers.tracking
        []
        >>> layers.add(1., age=1.)
        >>> layers.tracking
        ['age']
        """
        return [name for name in self._attrs if not name.startswith('_')]

    def _setup_layers(self, **kwds):
        dims = (self.allocated, self.number_of_stacks)
        for name, array in kwds.items():
            self._attrs[name] = _allocate_layers_for(array, *dims)

    @property
    def number_of_stacks(self):
        """Number of stacks."""
        return self._number_of_stacks

    @property
    def thickness(self):
        """Total thickness of the columns.

        The sum of all layer thicknesses for each stack as an array
        of shape `(number_of_stacks, )`.

        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers

        Initially there are no layers so the total thickness is 0.

        >>> layers = EventLayers(3)
        >>> layers.thickness
        array([ 0.,  0.,  0.])

        After adding some layers, the stacks have varying thicknesses.

        >>> layers.add(15.)
        >>> layers.add([1., -1., 2.])
        >>> layers.thickness
        array([ 16.,  14.,  17.])
        """
        return np.sum(self.dz, axis=0)

    @property
    def z(self):
        """Thickness to top of each layer.

        Thickness from the bottom of each stack to the top of each layer
        as an array of shape `(number_of_layers, number_of_stacks)`.
        
        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers

        Initially there are no layers so the elevation to the top
        is 0.

        >>> layers = EventLayers(3)
        >>> layers.z.shape == (0, 3)
        True

        After adding some layers, elevations are to the top of each layer.

        >>> layers.add(15.)
        >>> layers.add([1., -1., 2.])
        >>> layers.dz
        array([[ 15.,  14.,  15.],
               [  1.,   0.,   2.]])
        >>> layers.z
        array([[ 15.,  14.,  15.],
               [ 16.,  14.,  17.]])
        """
        return np.cumsum(self.dz, axis=0)

    @property
    def dz(self):
        """Thickness of each layer.

        The thickness of each layer at each stack as an array of shape
        `(number_of_layers, number_of_stacks)`.

        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers

        Initially there are no layers so there are not thicknesses.

        >>> layers = EventLayers(3)
        >>> layers.dz.shape == (0, 3)
        True

        Now add two layers, the first of uniform thickness and the
        second non-uniform and with some erosion.

        >>> layers.add(15.)
        >>> layers.add([1., -1., 2.])
        >>> layers.dz
        array([[ 15.,  14.,  15.],
               [  1.,   0.,   2.]])
        """
        return self._attrs['_dz'][:self.number_of_layers]

    @property
    def number_of_layers(self):
        """Total number of layers.

        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers

        >>> layers = EventLayers(3)
        >>> layers.number_of_layers
        0

        >>> layers.add(15.)
        >>> layers.add([1., -1., 2.])
        >>> layers.number_of_layers
        2
        """
        return self._number_of_layers

    @property
    def allocated(self):
        """Total number of allocated layers.

        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers

        >>> layers = EventLayers(3)
        >>> layers.number_of_layers
        0
        >>> layers.allocated == 0
        True

        >>> layers.add(15.)
        >>> layers.number_of_layers
        1
        >>> layers.allocated == 7
        True
        >>> for _ in range(layers.allocated): layers.add(0.)
        >>> layers.number_of_layers
        8
        >>> layers.allocated == 15
        True

        If you know how many layers you will ultimately have, you
        can allocated enough memory for them when you create your
        layer stacks.

        >>> layers = EventLayers(3, allocated=15)
        >>> layers.number_of_layers
        0
        >>> layers.allocated == 15
        True

        >>> layers.add(15.)
        >>> layers.number_of_layers
        1
        >>> layers.allocated == 15
        True
        >>> for _ in range(layers.allocated): layers.add(0.)
        >>> layers.number_of_layers
        16
        >>> layers.allocated == 24
        True
        """
        return self._attrs['_dz'].shape[0]

    def add(self, dz, **kwds):
        """Add a layer to the stacks.

        Parameters
        ----------
        dz : float or array_like
            Thickness to add to each stack.

        Examples
        --------
        >>> from landlab.layers.eventlayers import EventLayers

        Create an empty layer stack with 3 stacks.

        >>> layers = EventLayers(3)
        >>> layers.number_of_layers
        0

        To add a layer of uniform thickness to every stack.

        >>> layers.add(1.5)
        >>> layers.number_of_layers
        1
        >>> layers.dz
        array([[ 1.5,  1.5,  1.5]])

        Add a second layer with uneven thickness.

        >>> layers.add([1., 2., .5])
        >>> layers.dz
        array([[ 1.5,  1.5,  1.5],
               [ 1. ,  2. ,  0.5]])

        Adding a layer with negative thickness will remove
        existing layers for the top of the stack.

        >>> layers.add(-1)
        >>> layers.dz
        array([[ 1.5,  1.5,  1. ],
               [ 0. ,  1. ,  0. ],
               [ 0. ,  0. ,  0. ]])

        Use keywords to track properties of each layer. For instance,
        here we create a new stack and add a layer with a particular
        *age*. You can access the layer properties as if the object
        were a dictionary.

        >>> layers = EventLayers(3)
        >>> layers.add(1., age=3.)
        >>> layers.dz
        array([[ 1.,  1.,  1.]])
        >>> layers['age']
        array([[ 3.,  3.,  3.]])
        >>> layers.add(2., age=6.)
        >>> layers['age']
        array([[ 3.,  3.,  3.],
               [ 6.,  6.,  6.]])
        """
        if self.number_of_layers == 0:
            self._setup_layers(**kwds)

        self._add_empty_layer()

        _deposit_or_erode(self._attrs['_dz'], self.number_of_layers, dz)

        for name in kwds:
            try:
                self[name][-1] = kwds[name]
            except KeyError:
                print('{0} is not being tracked. Ignoring'.format(name),
                      file=sys.stderr)

    def _add_empty_layer(self):
        """Add a new empty layer to the stacks."""
        if self.number_of_layers >= self.allocated:
            self._resize(self.allocated + 1)

        self._number_of_layers += 1
        for name in self._attrs:
            self._attrs[name][self.number_of_layers - 1] = 0.

    def _resize(self, newsize, exact=False):
        """Allocate more memory for the layers."""
        for name in self._attrs:
            self._attrs[name] = resize_array(self._attrs[name], newsize,
                                             exact=exact)



#! /usr/env/python
""" Simple raster Landlab cellular automaton.

Simple raster Landlab cellular automaton, with
cell-pair transitions that depend on orientation (vertical or horizontal)

This file defines the OrientedRasterCTS class, which is a sub-class of
CellLabCTSModel that implements a simple, oriented, raster-grid
CA. Like its parent class, OrientedRasterCTS implements a continuous-time,
stochastic, pair-based CA.

Created GT Sep 2014
"""
from __future__ import print_function

import numpy as np

from .celllab_cts import CellLabCTSModel
from ..grid import RasterModelGrid


_DEBUG = False


class OrientedRasterCTS(CellLabCTSModel):

    """Oriented raster CellLab-CTS model.

    RasterCTS constructor: sets number of orientations to 2 and calls
    base-class constructor.

    Parameters
    ----------
    model_grid : Landlab ModelGrid object
        Reference to the model's grid
    node_state_dict : dict
        Keys are node-state codes, values are the names associated with
        these codes
    transition_list : list of Transition objects
        List of all possible transitions in the model
    initial_node_states : array of ints (x number of nodes in grid)
        Starting values for node-state grid
    prop_data : array (x number of nodes in grid) (optional)
        Array of properties associated with each node/cell
    prop_reset_value : number or object, optional
        Default or initial value for a node/cell property (e.g., 0.0).
        Must be same type as *prop_data*.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.ca.celllab_cts import Transition
    >>> from landlab.ca.oriented_raster_cts import OrientedRasterCTS

    >>> mg = RasterModelGrid(3, 4, 1.0)
    >>> nsd = {0 : 'yes', 1 : 'no'}
    >>> xnlist = []
    >>> xnlist.append(Transition((0,1,0), (1,1,0), 1.0, 'frogging'))
    >>> nsg = mg.add_zeros('node', 'node_state_grid')
    >>> orcts = OrientedRasterCTS(mg, nsd, xnlist, nsg)
    """

    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, prop_data=None, prop_reset_value=None):
        """
        RasterCTS constructor: sets number of orientations to 2 and calls
        base-class constructor.

        Parameters
        ----------
        model_grid : Landlab ModelGrid object
            Reference to the model's grid
        node_state_dict : dict
            Keys are node-state codes, values are the names associated with
            these codes
        transition_list : list of Transition objects
            List of all possible transitions in the model
        initial_node_states : array of ints (x number of nodes in grid)
            Starting values for node-state grid
        prop_data : array (x number of nodes in grid) (optional)
            Array of properties associated with each node/cell
        prop_reset_value : number or object, optional
            Default or initial value for a node/cell property (e.g., 0.0).
            Must be same type as *prop_data*.
        """

        if _DEBUG:
            print('OrientedRasterCTS.__init__ here')

        # Make sure caller has sent the right grid type
        if not isinstance(model_grid, RasterModelGrid):
            raise TypeError('model_grid must be a Landlab RasterModelGrid')

        # Define the number of distinct cell-pair orientations: here just 1,
        # because RasterLCA represents a non-oriented CA model.
        self.number_of_orientations = 2

        # Call the LandlabCellularAutomaton constructor to do the rest of
        # the initialization
        super(OrientedRasterCTS, self).__init__(model_grid, node_state_dict,
                                                transition_list,
                                                initial_node_states, prop_data,
                                                prop_reset_value)

        if _DEBUG:
            print('ORCTS:')
            print(self.n_xn)
            print(self.xn_to)
            print(self.xn_rate)

    def setup_array_of_orientation_codes(self):
        """
        Creates and configures an array that contain the orientation code for
        each active link (and corresponding cell pair).

        Notes
        -----
        **Creates**:

        * ``self.active_link_orientation``: 1D numpy array of ints
          Array of orientation codes for each cell pair (link)

        This overrides the method of the same name in landlab_ca.py.
        """
        # Create array for the orientation of each active link
        self.link_orientation = np.zeros(self.grid.number_of_links,
                                         dtype=np.int8)

        # Set its value according to the different in y coordinate between each
        # link's TO and FROM nodes (the numpy "astype" method turns the
        # resulting array into integer format)
        dy = (self.grid.node_y[self.grid.node_at_link_head] -
              self.grid.node_y[self.grid.node_at_link_tail])
        self.link_orientation = dy.astype(np.int8)

        if _DEBUG:
            print(self.link_orientation)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#! /usr/bin/env python
"""
Unit tests for landlab.io.esri_ascii module.
"""
import os

import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
from six import StringIO

from landlab.io import read_esri_ascii, read_asc_header
from landlab.io import (MissingRequiredKeyError, KeyTypeError, DataSizeError,
                        BadHeaderLineError, KeyValueError, 
                        MismatchGridDataSizeError)
from landlab import RasterModelGrid


_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


def test_hugo_read_file_name():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR,
                                                 'hugo_site.asc'))

    assert isinstance(grid, RasterModelGrid)

    assert field.size == 55 * 76
    assert field.shape == (55 * 76, )


def test_hugo_read_file_like():
    with open(os.path.join(_TEST_DATA_DIR, 'hugo_site.asc')) as asc_file:
        (grid, field) = read_esri_ascii(asc_file)

    assert isinstance(grid, RasterModelGrid)

    assert field.size == 55 * 76
    assert field.shape == (55 * 76, )


def test_hugo_reshape():
    with open(os.path.join(_TEST_DATA_DIR, 'hugo_site.asc')) as asc_file:
        (grid, field) = read_esri_ascii(asc_file, reshape=True)

    assert isinstance(grid, RasterModelGrid)

    assert field.shape == (55, 76)


def test_4x3_read_file_name():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR,
                                                 '4_x_3.asc'))

    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(field,
                       np.array([9., 10., 11.,
                                 6.,  7.,  8.,
                                 3.,  4.,  5.,
                                 0.,  1.,  2.]))


def test_4x3_read_file_like():
    with open(os.path.join(_TEST_DATA_DIR, '4_x_3.asc')) as asc_file:
        (grid, field) = read_esri_ascii(asc_file)

    assert isinstance(grid, RasterModelGrid)

    assert_array_equal(field,
                       np.array([9., 10., 11.,
                                 6.,  7.,  8.,
                                 3.,  4.,  5.,
                                 0.,  1.,  2.]))


def test_4x3_shape_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4.
5. 6. 7. 8.
9. 10. 11. 12.
        """)
    (grid, field) = read_esri_ascii(asc_file)
    assert field.size == 12

    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
        """)
    (grid, field) = read_esri_ascii(asc_file)
    assert field.size == 12


def test_4x3_size_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10.
        """)
    with pytest.raises(DataSizeError):
        read_esri_ascii(asc_file)
    
def test_grid_data_size_mismatch():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
1. 2. 3. 4. 5. 6. 7. 8. 9. 10. 11. 12.
        """)
    rmg = RasterModelGrid((10,10),10.)
    with pytest.raises(MismatchGridDataSizeError):
        read_esri_ascii(asc_file, grid=rmg)    


def test_header_missing_required_key():
    asc_file = StringIO(
        """
nrows         4
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """)
    with pytest.raises(MissingRequiredKeyError):
        read_asc_header(asc_file)


def test_header_unknown_key():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
invalid_key   1
        """)
    with pytest.raises(BadHeaderLineError):
        read_asc_header(asc_file)


def test_header_missing_value():
    asc_file = StringIO(
        """
nrows         4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize
NODATA_value  -9999
invalid_key   1
        """)
    with pytest.raises(BadHeaderLineError):
        read_asc_header(asc_file)


def test_header_bad_values():
    asc_file = StringIO(
        """
nrows         -4
ncols         3
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """)
    with pytest.raises(KeyValueError):
        read_asc_header(asc_file)


def test_header_missing_mutex_key():
    asc_file = StringIO(
        """
ncols         3
nrows         4
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """)
    with pytest.raises(MissingRequiredKeyError):
        read_asc_header(asc_file)


def test_header_mutex_key():
    asc_file = StringIO(
        """
ncols         3
nrows         4
xllcenter     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """)
    header = read_asc_header(asc_file)
    assert header['xllcenter'] == 1.
    with pytest.raises(KeyError):
        header['xllcorner']

    asc_file = StringIO(
        """
ncols         3
nrows         4
xllcorner     1.
yllcorner     2.
cellsize      10.
NODATA_value  -9999
        """)
    header = read_asc_header(asc_file)
    assert header['xllcorner'] == 1.
    with pytest.raises(KeyError):
        header['xllcenter']


def test_header_missing_optional():
    asc_file = StringIO(
        """
ncols         3
nrows         4
xllcenter     1.
yllcorner     2.
cellsize      10.
        """)
    header = read_asc_header(asc_file)
    with pytest.raises(KeyError):
        header['nodata_value']


def test_header_case_insensitive():
    asc_file = StringIO(
        """
nCoLs         3
nrows         4
Xllcenter     1.
YLLCORNER     2.
CELLSIZE      10.
NODATA_value  -999
        """)
    header = read_asc_header(asc_file)
    for key in ['ncols', 'nrows', 'xllcenter', 'yllcorner', 'cellsize',
                'nodata_value']:
        assert key in header


def test_header_wrong_type():
    asc_file = StringIO(
        """
nCoLs         3.5
nrows         4
Xllcenter     1.
YLLCORNER     2.
CELLSIZE      10.
NODATA_value  -999
        """)
    with pytest.raises(KeyTypeError):
        read_asc_header(asc_file)


def test_name_keyword():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR,
                                                 '4_x_3.asc'),
                                    name='air__temperature')

    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(field,
                       np.array([9., 10., 11.,
                                 6.,  7.,  8.,
                                 3.,  4.,  5.,
                                 0.,  1.,  2.]))
    assert_array_almost_equal(grid.at_node['air__temperature'], field)
    assert grid.at_node['air__temperature'] is field
    
def test_halo_keyword():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR, \
                                                 '4_x_3.asc'), \
                                                 halo=1)
                                    
    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(field,
                       np.array([-9999., -9999., -9999., -9999., -9999.,  
                                 -9999.,     9.,    10.,    11., -9999.,
                                 -9999.,     6.,     7.,     8., -9999.,
                                 -9999.,     3.,     4.,     5., -9999.,
                                 -9999.,     0.,     1.,     2., -9999.,
                                 -9999., -9999., -9999., -9999., -9999.]))
                                 
def test_halo_keyword_no_nodata_value():
    (grid, field) = read_esri_ascii(os.path.join(_TEST_DATA_DIR, \
                                                '4_x_3_no_nodata_value.asc'), \
                                                halo=1)
                                    
    assert isinstance(grid, RasterModelGrid)

    assert isinstance(field, np.ndarray)
    assert_array_equal(field,
                       np.array([-9999., -9999., -9999., -9999., -9999.,  
                                 -9999.,     9.,    10.,    11., -9999.,
                                 -9999.,     6.,     7.,     8., -9999.,
                                 -9999.,     3.,     4.,     5., -9999.,
                                 -9999.,     0.,     1.,     2., -9999.,
                                 -9999., -9999., -9999., -9999., -9999.]))


if __name__ == '__main__':
    unittest.main()



"""
Unit tests for landlab.components.radiation.radiation
"""
import pytest
from numpy.testing import assert_array_almost_equal
import numpy as np


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(rad):
    assert rad.name == 'Radiation'


def test_input_var_names(rad):
    assert rad.input_var_names == ('topographic__elevation',)


def test_output_var_names(rad):
    assert sorted(rad.output_var_names) == [
        'radiation__incoming_shortwave_flux',
        'radiation__net_shortwave_flux',
        'radiation__ratio_to_flat_surface'
    ]


def test_var_units(rad):
    assert set(rad.input_var_names) | set(rad.output_var_names) == set(dict(rad.units).keys())

    assert rad.var_units('topographic__elevation') == 'm'
    assert rad.var_units('radiation__incoming_shortwave_flux') == 'W/m^2'
    assert rad.var_units('radiation__net_shortwave_flux') == 'W/m^2'
    assert rad.var_units('radiation__ratio_to_flat_surface') == 'None'


def test_grid_shape(rad):
    assert rad.grid.number_of_node_rows == _SHAPE[0]
    assert rad.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(rad):
    assert rad.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(rad):
    assert rad.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(rad):
    for name in rad.grid['node']:
        field = rad.grid['node'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (rad.grid.number_of_node_rows * rad.grid.number_of_node_columns, )
                      
    for name in rad.grid['cell']:
        field = rad.grid['cell'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (rad.grid.number_of_cell_rows * rad.grid.number_of_cell_columns, )

    with pytest.raises(KeyError):
        rad.grid['not_a_var_name']


def test_field_initialized_to_zero(rad):
    for name in rad.grid['node']:
        field = rad.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(rad.grid.number_of_nodes))
    for name in rad.grid['cell']:
        if name == 'Slope' or name == 'Aspect':
            continue
        field = rad.grid['cell'][name]
        assert_array_almost_equal(field, np.zeros(rad.grid.number_of_cells))



from .flow_accum_bw import (make_ordered_node_array,
                            find_drainage_area_and_discharge,
                            flow_accumulation)

from .flow_accumulator import FlowAccumulator

__all__ = ['FlowAccumulator',
           'make_ordered_node_array', 'find_drainage_area_and_discharge',
           'flow_accumulation', ]



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 10:48:23 2015

@author: gtucker
"""

from landlab import RasterModelGrid
from numpy.testing import assert_array_equal


def test_link_order():
    """Test ordering of links."""
    rg = RasterModelGrid(3, 4)
    assert_array_equal(rg.node_at_link_tail, [0, 1, 2, 0, 1, 2,  3, 4, 5, 6,
                                              4, 5, 6, 7, 8, 9, 10])
    assert_array_equal(rg.node_at_link_head, [1, 2,  3,  4, 5,  6,  7, 5, 6, 7,
                                              8, 9, 10, 11, 9, 10, 11])


if __name__ == '__main__':
    test_link_order()




from landlab import Component
import numpy as np
from ...utils.decorators import use_file_name_or_kwds

_VALID_METHODS = set(['Grid'])
GRASS = 0
SHRUB = 1
TREE = 2
BARE = 3
SHRUBSEEDLING = 4
TREESEEDLING = 5


def assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class VegCA(Component):
    """
    Landlab component that simulates inter-species plant competition using
    a 2D cellular automata model.

    This code is based on Cellular Automata Tree Grass Shrub Simulator (CATGraSS).
    It simulates spatial competition of multiple plant functional types through
    establishment and mortality. In the current code, tree, grass and
    shrubs are used.

    Ref: Zhou, X., Istanbulluoglu, E., & Vivoni, E. R. (2013). Modeling the
    ecohydrological role of aspect controlled radiation on tree grass shrub
    coexistence in a semiarid climate. Water Resources Research,
    49(5), 2872-2895.

    .. codeauthor:: Sai Nudurupati and Erkan Istanbulluoglu

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import VegCA
    >>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))
    >>> VegCA.name
    'Cellular Automata Plant Competition'
    >>> sorted(VegCA.output_var_names)
    ['plant__age', 'plant__live_index']
    >>> sorted(VegCA.units) # doctest: +NORMALIZE_WHITESPACE
    [('plant__age', 'Years'),
     ('plant__live_index', 'None'),
     ('vegetation__cumulative_water_stress', 'None'),
     ('vegetation__plant_functional_type', 'None')]
    >>> grid['cell']['vegetation__plant_functional_type']= np.arange(
    ...        0, grid.number_of_cells, dtype=int)
    >>> grid['cell']['vegetation__cumulative_water_stress'] = np.ones(
    ...        grid.number_of_cells)
    >>> ca_veg = VegCA(grid)
    >>> ca_veg.grid.number_of_cell_rows
    3
    >>> ca_veg.grid.number_of_cell_columns
    2
    >>> ca_veg.grid is grid
    True
    >>> import numpy as np
    >>> A = np.copy(grid['cell']['plant__age'])
    >>> ca_veg.update()
    >>> np.alltrue(grid['cell']['plant__age'] == A)
    False
    """
    _name = 'Cellular Automata Plant Competition'

    _input_var_names = (
        'vegetation__cumulative_water_stress',
        'vegetation__plant_functional_type',
    )

    _output_var_names = (
        'plant__live_index',
        'plant__age',
    )

    _var_units = {
        'vegetation__cumulative_water_stress': 'None',
        'vegetation__plant_functional_type': 'None',
        'plant__live_index': 'None',
        'plant__age': 'Years',
    }

    _var_mapping = {
        'vegetation__cumulative_water_stress': 'cell',
        'vegetation__plant_functional_type': 'cell',
        'plant__live_index': 'cell',
        'plant__age': 'cell',
    }

    _var_doc = {
        'vegetation__cumulative_water_stress':
            'cumulative vegetation__water_stress over the growing season',
        'vegetation__plant_functional_type':
            'classification of plants (int), grass=0, shrub=1, tree=2, \
             bare=3, shrub_seedling=4, tree_seedling=5',
        'plant__live_index':
            '1 - vegetation__cumulative_water_stress',
        'plant__age':
            'Age of plant',
    }

    @use_file_name_or_kwds
    def __init__(self, grid,
                 Pemaxg=0.35, ING=2., ThetaGrass=0.62, PmbGrass=0.05,
                 Pemaxsh=0.2, ThetaShrub=0.8, PmbShrub=0.01, tpmaxShrub=600,
                 Pemaxtr=0.25, ThetaTree=0.72, PmbTree=0.01, tpmaxTree=350,
                 ThetaShrubSeedling=0.64, PmbShrubSeedling=0.03,
                 tpmaxShrubSeedling=18, ThetaTreeSeedling=0.64,
                 PmbTreeSeedling=0.03, tpmaxTreeSeedling=18,
                 **kwds):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        Pemaxg: float, optional
            Maximal establishment probability of grass.
        ING: float, optional
            Parameter to define allelopathic effect of creosote on grass.
        ThetaGrass: float, optional
            Drought resistance threshold of grass.
        PmbGrass: float, optional
            Background mortality probability of grass.
        Pemaxsh: float, optional
            Maximal establishment probability of shrub.
        ThetaShrub: float, optional
            Drought resistance threshold of shrub.
        PmbShrub: float, optional
            Background mortality probability of shrub.
        tpmaxShrub: float, optional
            Maximum age of shrub (years).
        Pemaxtr: float, optional
            Maximal establishment probability of tree.
        Thetatree: float, optional
            Drought resistance threshold of tree.
        PmbTree: float, optional
            Background mortality probability of tree.
        tpmaxTree: float, optional
            Maximum age of tree (years).
        ThetaShrubSeedling: float, optional
            Drought resistance threshold of shrub seedling.
        PmbShrubSeedling: float, optional
            Background mortality probability of shrub seedling.
        tpmaxShrubSeedling: float, optional
            Maximum age of shrub seedling (years).
        ThetaTreeSeedling: float, optional
            Drought resistance threshold of tree seedling.
        PmbTreeSeedling: float, optional
            Background mortality probability of tree seedling.
        tpmaxTreeSeedling: float, optional
            Maximum age of tree seedling (years).
        """
        self._method = kwds.pop('method', 'Grid')
        self._Pemaxg = Pemaxg   # Pe-max-grass - max probability
        self._Pemaxsh = Pemaxsh    # Pe-max-shrub
        self._Pemaxtr = Pemaxtr   # Pe-max-tree
        self._INg = ING  # Allelopathic effect on grass from creosotebush
        self._th_g = ThetaGrass  # grass
        self._th_sh = ThetaShrub  # shrub - Creosote
        self._th_tr = ThetaTree  # Juniper pine
        self._th_sh_s = ThetaShrubSeedling  # shrub seedling
        self._th_tr_s = ThetaTreeSeedling  # Juniper pine seedling
        self._Pmb_g = PmbGrass  # Background mortality probability - grass
        self._Pmb_sh = PmbShrub  # shrub
        self._Pmb_tr = PmbTree  # tree
        self._Pmb_sh_s = PmbShrubSeedling  # shrub seedling
        self._Pmb_tr_s = PmbTreeSeedling  # tree seedling
        self._tpmax_sh = tpmaxShrub  # Maximum age - shrub
        self._tpmax_tr = tpmaxTree  # Maximum age - tree
        self._tpmax_sh_s = tpmaxShrubSeedling  # Maximum age - shrub seedling
        self._tpmax_tr_s = tpmaxTreeSeedling  # Maximum age - tree seedling

        assert_method_is_valid(self._method)

        super(VegCA, self).__init__(grid)

        if 'vegetation__plant_functional_type' not in self.grid.at_cell:
            grid['cell']['vegetation__plant_functional_type'] = (
                    np.random.randint(0, 6, grid.number_of_cells))

        for name in self._input_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        self._cell_values = self.grid['cell']

        VegType = grid['cell']['vegetation__plant_functional_type']
        tp = np.zeros(grid.number_of_cells, dtype=int)
        tp[VegType == TREE] = np.random.randint(
            0, self._tpmax_tr, np.where(VegType == TREE)[0].shape)
        tp[VegType == SHRUB] = np.random.randint(
            0, self._tpmax_sh, np.where(VegType == SHRUB)[0].shape)
        locs_trees = np.where(VegType == TREE)[0]
        locs_shrubs = np.where(VegType == SHRUB)[0]
        VegType[locs_trees[tp[locs_trees] < self._tpmax_tr_s]] = TREESEEDLING
        VegType[locs_shrubs[tp[
            locs_shrubs] < self._tpmax_sh_s]] = SHRUBSEEDLING
        grid['cell']['plant__age'] = tp

    def update(self, time_elapsed=1, Edit_VegCov=True):
        """
        Update fields with current loading conditions.

        Parameters
        ----------
        time_elapsed: int, optional
            Time elapsed - time step (years).
        Edit_VegCov: switch (0 or 1), optional
            If Edit_VegCov=1, an optional field 'vegetation__boolean_vegetated'
            will be output, (i.e.) if a cell is vegetated the corresponding
            cell of the field will be 1, otherwise it will be 0.
        """
        self._VegType = self._cell_values['vegetation__plant_functional_type']
        self._CumWS = self._cell_values[
            'vegetation__cumulative_water_stress']
        self._live_index = self._cell_values['plant__live_index']
        self._tp = self._cell_values['plant__age'] + time_elapsed

        # Check if shrub and tree seedlings have matured
        shrub_seedlings = np.where(self._VegType == SHRUBSEEDLING)[0]
        tree_seedlings = np.where(self._VegType == TREESEEDLING)[0]
        matured_shrubs = np.where(self._tp[shrub_seedlings] >
                                  self._tpmax_sh_s)[0]
        matured_trees = np.where(self._tp[tree_seedlings] >
                                 self._tpmax_tr_s)[0]
        self._VegType[shrub_seedlings[matured_shrubs]] = SHRUB
        self._VegType[tree_seedlings[matured_trees]] = TREE
        self._tp[shrub_seedlings[matured_shrubs]] = 0
        self._tp[tree_seedlings[matured_trees]] = 0

        # Establishment
        self._live_index = 1 - self._CumWS      # Plant live index = 1 - WS
        bare_cells = np.where(self._VegType == BARE)[0]
        n_bare = len(bare_cells)
        first_ring = self.grid.looped_neighbors_at_cell[bare_cells]
        second_ring =                                                   \
            self.grid.second_ring_looped_neighbors_at_cell[bare_cells]
        veg_type_fr = self._VegType[first_ring]
        veg_type_sr = self._VegType[second_ring]
        Sh_WS_fr = WS_PFT(veg_type_fr, SHRUB, self._live_index[first_ring])
        Tr_WS_fr = WS_PFT(veg_type_fr, TREE, self._live_index[first_ring])
        Tr_WS_sr = WS_PFT(veg_type_sr, TREE, self._live_index[second_ring])

        n = count(veg_type_fr, SHRUB)
        Phi_sh = Sh_WS_fr/8.
        Phi_tr = (Tr_WS_fr + Tr_WS_sr/2.)/8.
        Phi_g = np.mean(self._live_index[np.where(self._VegType == GRASS)])
        Pemaxg = self._Pemaxg * np.ones(n_bare)
        Pemaxsh = self._Pemaxsh * np.ones(n_bare)
        Pemaxtr = self._Pemaxtr * np.ones(n_bare)
        Peg = np.amin(np.vstack((Phi_g/(n*self._INg), Pemaxg)), axis=0)
        Pesh = np.amin(np.vstack((Phi_sh, Pemaxsh)), axis=0)
        Petr = np.amin(np.vstack((Phi_tr, Pemaxtr)), axis=0)
        Select_PFT_E = np.random.choice([GRASS, SHRUBSEEDLING, TREESEEDLING],
                                        n_bare)
        # Grass - 0; Shrub Seedling - 4; Tree Seedling - 5
        Pest = np.choose(Select_PFT_E, [Peg, 0, 0, 0, Pesh, Petr])
        # Probability of establishment
        R_Est = np.random.rand(n_bare)
        # Random number for comparison to establish
        Establish = np.int32(np.where(
                             np.greater_equal(Pest, R_Est) == True)[0])
        self._VegType[bare_cells[Establish]] = Select_PFT_E[Establish]
        self._tp[bare_cells[Establish]] = 0

        # Mortality
        plant_cells = np.where(self._VegType != BARE)[0]
        n_plant = len(plant_cells)
        Theta = np.choose(self._VegType[plant_cells], [
                          self._th_g, self._th_sh, self._th_tr,
                          0, self._th_sh_s, self._th_tr_s])
        PMd = self._CumWS[plant_cells] - Theta
        PMd[PMd < 0.] = 0.
        tpmax = np.choose(self._VegType[plant_cells], [
                          200000, self._tpmax_sh, self._tpmax_tr,
                          0, self._tpmax_sh_s, self._tpmax_tr_s])
        PMa = np.zeros(n_plant)
        tp_plant = self._tp[plant_cells]
        tp_greater = np.where(tp_plant > 0.5*tpmax)[0]
        PMa[tp_greater] = ((tp_plant[tp_greater] - 0.5 * tpmax[tp_greater]) /
                           (0.5 * tpmax[tp_greater])) - 1
        PMb = np.choose(self._VegType[plant_cells], [
                        self._Pmb_g, self._Pmb_sh, self._Pmb_tr, 0,
                        self._Pmb_sh_s, self._Pmb_tr_s])
        PM = PMd + PMa + PMb
        PM[PM > 1.] = 1.
        R_Mor = np.random.rand(n_plant)  # Random number for comparison to kill
        Mortality = np.int32(np.where(np.greater_equal(PM, R_Mor) == True)[0])
        self._VegType[plant_cells[Mortality]] = BARE
        self._tp[plant_cells[Mortality]] = 0

        self._cell_values['plant__age'] = self._tp

        if Edit_VegCov:
            self.grid['cell']['vegetation__boolean_vegetated'] =             \
                        np.zeros(self.grid.number_of_cells, dtype=int)
            self.grid['cell']['vegetation__boolean_vegetated'][
                        self._VegType != BARE] = 1

        # For debugging purposes
        self._bare_cells = bare_cells
        self._Established = bare_cells[Establish]
        self._plant_cells = plant_cells
        self._Mortified = plant_cells[Mortality]


def count(Arr, value):
    Res = np.zeros(Arr.shape[0], dtype=int)
    x, y = Arr.shape
    for i in range(0, x):
        for j in range(0, y):
            if Arr[i][j] == value:
                Res[i] += 1
    return Res


def WS_PFT(VegType, PlantType, WS):
    Phi = np.zeros(WS.shape[0])
    x, y = WS.shape
    for i in range(0, x):
        for j in range(0, y):
            if VegType[i][j] == PlantType:
                Phi[i] += WS[i][j]
    return Phi



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_links_at_patch(shape, np.ndarray[DTYPE_t, ndim=2] links_at_patch):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int links_per_row = 2 * n_cols - 1
    cdef int patches_per_row = n_cols - 1

    for row in range(n_rows - 1):
        link = row * links_per_row + n_cols
        patch = row * patches_per_row
        for col in range(n_cols - 1):
            links_at_patch[patch, 0] = link
            links_at_patch[patch, 1] = link + n_cols - 1
            links_at_patch[patch, 2] = link - 1
            links_at_patch[patch, 3] = link - n_cols

            patch += 1
            link += 1



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid


def test_unit_grid_all_cells():
    rmg = RasterModelGrid((4, 4))
    assert_array_equal(rmg.area_of_cell, np.ones(4))


def test_unit_grid_one_cell():
    rmg = RasterModelGrid((4, 4))
    assert_array_equal(rmg.area_of_cell[0], 1.)


def test_unit_grid_last_cell():
    rmg = RasterModelGrid((4, 4))
    assert_array_equal(rmg.area_of_cell[-1], 1.)


def test_out_of_range():
    rmg = RasterModelGrid((4, 4))
    with pytest.raises(IndexError):
        rmg.area_of_cell[5]


def test_is_immutable():
    rmg = RasterModelGrid((4, 4))
    with pytest.raises(ValueError):
        rmg.area_of_cell[0] = 0.


def test_all_cells_with_spacing():
    rmg = RasterModelGrid(4, 4, 10.)
    assert_array_equal(rmg.area_of_cell, 100 * np.ones(4))



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab.grid import raster_funcs as rfuncs


def test_best_fit_with_bad_args():
    """Raise an error if input arrays are not the same length."""
    len_2 = np.array([0, 1])
    len_3 = np.array([0, 0, 1])

    with pytest.raises(ValueError):
        rfuncs.calculate_slope_aspect_bfp(len_2, len_3, len_3)
    with pytest.raises(ValueError):
        rfuncs.calculate_slope_aspect_bfp(len_3, len_2, len_3)
    with pytest.raises(ValueError):
        rfuncs.calculate_slope_aspect_bfp(len_3, len_3, len_2)


def test_best_fit_plane_in_xy():
    """Best fit plane is the xy-plane."""
    (x, y, z) = (np.array([0., 1., 1.]),
                 np.array([0., 0., 1.]),
                 np.array([0., 0., 0.]), )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 0.
    assert aspect == 90.

    (x, y, z) = (np.array([0., 1., 1., 0.]),
                 np.array([0., 0., 1., 1.]),
                 np.array([0., 0., 0., 0.]), )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 0.
    assert aspect == 90.


def test_best_fit_plane_in_xz():
    """Best fit plane is the xz-plane."""
    (x, y, z) = (np.array([0., 1., 1.]),
                 np.array([0., 0., 0.]),
                 np.array([0., 0., 1.]), )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 90.
    assert aspect == 0.


def test_best_fit_plane_in_yz():
    """Best fit plane is the yz-plane."""
    (x, y, z) = (np.array([0., 0., 0.]),
                 np.array([0., 1., 1.]),
                 np.array([0., 0., 1.]), )
    (slope, aspect) = rfuncs.calculate_slope_aspect_bfp(x, y, z)

    assert slope == 90.
    assert aspect == 90.



#! /usr/bin/env python
import numpy as np
import pytest

from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.grid.diagonals import create_nodes_at_diagonal


def test_nodes_at_diagonal():
    """Test tail and head nodes of diagonals."""
    diagonals = create_nodes_at_diagonal((2, 2))
    assert_array_equal(diagonals, [[0, 3], [1, 2]])

    diagonals = create_nodes_at_diagonal((4, 3))
    assert isinstance(diagonals, np.ndarray)
    assert_array_equal(
        diagonals,
        [
            [0, 4],
            [1, 3],
            [1, 5],
            [2, 4],
            [3, 7],
            [4, 6],
            [4, 8],
            [5, 7],
            [6, 10],
            [7, 9],
            [7, 11],
            [8, 10],
        ],
    )


def test_nodes_at_diagonal_1d():
    """Test nodes at diagonals for 1d grid."""
    diagonals = create_nodes_at_diagonal((1, 2))
    assert_array_equal(diagonals, np.array([], dtype=int).reshape((0, 2)))

    diagonals = create_nodes_at_diagonal((4, 1))
    assert_array_equal(diagonals, np.array([], dtype=int).reshape((0, 2)))


def test_nodes_at_diagonal_out_keyword():
    """Test out keyword for nodes_at_diagonal."""
    buffer = np.empty((4, 2), dtype=int)
    diagonals = create_nodes_at_diagonal((3, 2), out=buffer)

    assert buffer is diagonals
    assert_array_equal(diagonals, [[0, 3], [1, 2], [2, 5], [3, 4]])


def test_diagonal_values_are_cached(diagonal_property):
    grid = RasterModelGrid((4, 3))
    x = getattr(grid, diagonal_property)
    assert x is getattr(grid, diagonal_property)


def test_diagonal_values_are_readonly(diagonal_property):
    grid = RasterModelGrid((4, 3))
    x = getattr(grid, diagonal_property)
    with pytest.raises(ValueError):
        x[0] = 999
    assert x.flags["WRITEABLE"] is False



import pytest
from numpy.testing import assert_array_equal

from landlab.grid.unstructured.links import (link_ids_at_node,
                                             in_link_ids_at_node,
                                             out_link_ids_at_node)


def test_array_length_mismatch():
    with pytest.raises(ValueError):
        link_ids_at_node(([0, 1, 2, 3, 4, 5], [3, 4, 5, 6, 7]))


def test_link_arg_as_scalars():
    links, count = link_ids_at_node((0, 1), number_of_nodes=2)
    assert_array_equal(links, [0, 0])
    assert_array_equal(count, [1, 1])

    links, count = in_link_ids_at_node((0, 1), number_of_nodes=2)
    assert_array_equal(links, [0])
    assert_array_equal(count, [0, 1])

    links, count = out_link_ids_at_node((0, 1), number_of_nodes=2)
    assert_array_equal(links, [0])
    assert_array_equal(count, [1, 0])


def test_link_arg_as_pairs():
    links, count = link_ids_at_node(
        [(0, 1), (0, 2), (0, 3)], number_of_nodes=4)
    assert_array_equal(links, [0, 1, 2, 0, 1, 2])
    assert_array_equal(count, [3, 1, 1, 1])

    links, count = in_link_ids_at_node(
        [(0, 1), (0, 2), (0, 3)], number_of_nodes=4)
    assert_array_equal(links, [0, 1, 2])
    assert_array_equal(count, [0, 1, 1, 1])

    links, count = out_link_ids_at_node(
        [(0, 1), (0, 2), (0, 3)], number_of_nodes=4)
    assert_array_equal(links, [0, 1, 2])
    assert_array_equal(count, [3, 0, 0, 0])


def test_link_arg_as_tuple_sequence():
    links, count = link_ids_at_node([(0, 1), (1, 0)], number_of_nodes=2)
    assert_array_equal(links, [1, 0, 0, 1])
    assert_array_equal(count, [2, 2])

    links, count = in_link_ids_at_node([(0, 1), (1, 0)], number_of_nodes=2)
    assert_array_equal(links, [1, 0])
    assert_array_equal(count, [1, 1])

    links, count = out_link_ids_at_node([(0, 1), (1, 0)], number_of_nodes=2)
    assert_array_equal(links, [0, 1])
    assert_array_equal(count, [1, 1])


def test_links_3x3_quad():
    link_ends = [(0, 3), (1, 4), (2, 5), (3, 6), (4, 7), (5, 8),
                 (0, 1), (1, 2), (3, 4), (4, 5), (6, 7), (7, 8)]

    (links, count) = link_ids_at_node(list(zip(*link_ends)), number_of_nodes=9)

    assert_array_equal(links, [0, 6, 6, 1, 7, 7, 2,
                               0, 3, 8, 1, 8, 4, 9, 2, 9, 5,
                               3, 10, 4, 10, 11, 5, 11])
    assert_array_equal(count, [2, 3, 2, 3, 4, 3, 2, 3, 2])


def test_out_links_2x2_quad():
    link_ends = [(0, 2), (1, 3), (0, 1), (2, 3)]
    (links, count) = out_link_ids_at_node(zip(*link_ends), number_of_nodes=4)
    assert_array_equal(links, [0, 2, 1, 3])
    assert_array_equal(count, [2, 1, 1, 0])


def test_out_links_3x3_quad():
    link_ends = [(0, 3), (1, 4), (2, 5), (3, 6), (4, 7), (5, 8),
                 (0, 1), (1, 2), (3, 4), (4, 5), (6, 7), (7, 8)]

    (links, count) = out_link_ids_at_node(link_ends, number_of_nodes=9)
    assert_array_equal(links, [0, 6, 1, 7, 2, 3, 8, 4, 9, 5, 10, 11])
    assert_array_equal(count, [2, 2, 1, 2, 2, 1, 1, 1, 0])
